/*
 Copyright (C) JAEHYUK CHO
 All rights reserved.
 
 Author: JAEHYUK CHO <minzkn@minzkn.com>
 
 http://www.minzkn.com
*/

#if !defined(__def_mzproxy_header_mzproxy_h__)
# define __def_mzproxy_header_mzproxy_h__ "mzproxy.h"

/* ---- */

#if defined(HAVE_CONFIG_H)
# include "config.h"
#endif

#if defined(WIN32) || defined(_WIN32) || defined(WIN64) || defined(_WIN64)
# define def_mzproxy_windows                                         (1L)
#else
# define def_mzproxy_windows                                         (0L)
#endif

#if def_mzproxy_windows != (0L)
# define def_mzproxy_winsock                                         (1L)
# if !defined(ssize_t)
#  define ssize_t long
# endif
#else
# define def_mzproxy_winsock                                         (0L)
#endif

#if defined(__cplusplus)
# define def_mzproxy_cplusplus                                       (1L)
#else
# define def_mzproxy_cplusplus                                       (0L)
#endif

/* ---- */

#if def_mzproxy_cplusplus != (0L)
# define def_mzproxy_import                                          extern "C"
# define def_mzproxy_import_core                                     extern "C"
# define def_mzproxy_export                                          extern "C"
#else
# define def_mzproxy_import                                          extern
# define def_mzproxy_import_core                                     extern
# define def_mzproxy_export
#endif

/* ---- */

/* optimize cache control macro */
#if defined(__GNUC__)
# if __GNUC__ < (3L) /* no expect macro */
#  define mzproxy_builtin_expect(m_expression,m_value)               m_expression
#  define mzproxy_printf_varg_check(m_format_index,m_varg_index)
# else
#  define mzproxy_builtin_expect(m_expression,m_value)               __builtin_expect((long)(m_expression),(long)(m_value))
#  define mzproxy_printf_varg_check(m_format_index,m_varg_index)     __attribute__((__format__(__printf__,m_format_index,m_varg_index)))
# endif
#else
# define mzproxy_builtin_expect(m_expression,m_value)                m_expression
# define mzproxy_printf_varg_check(m_format_index,m_varg_index)
#endif

/* ---- */

#if !defined(__def_mzproxy_use_telnetd__)
# if def_mzproxy_windows != (0L)
#  define __def_mzproxy_use_telnetd__                                (0L)
# else
#  define __def_mzproxy_use_telnetd__                                (1L)
# endif
#endif

#if !defined(__def_mzproxy_use_ifaddrs__)
# if def_mzproxy_winsock != (0L)
#  define __def_mzproxy_use_ifaddrs__                                (0L)
# else
#  define __def_mzproxy_use_ifaddrs__                                (1L)
# endif
#endif

#if !defined(__def_mzproxy_use_inet6__)
# define __def_mzproxy_use_inet6__                                   (1L)
#endif

#if !defined(__def_mzproxy_use_socks__)
# define __def_mzproxy_use_socks__                                   (1L)
#endif

#if !defined(__def_mzproxy_use_syslog__)
# if def_mzproxy_windows != (0L)
#  define __def_mzproxy_use_syslog__                                 (0L)
#else
#  define __def_mzproxy_use_syslog__                                 (1L)
# endif
#endif

/* ---- */

#if def_mzproxy_windows != (0L)
# include <winsock2.h>
# include <ws2tcpip.h>
# include <windows.h>
# include <sys/types.h>
# include <io.h>
# include <direct.h> 
# include <stdio.h>
# include <stdlib.h>
# include <stdarg.h>
# include <memory.h>
# include <string.h>
# include <time.h>
# include <fcntl.h>
# include <errno.h>
# include <signal.h>
# define mzproxy_socket_t SOCKET
# define mzproxy_socklen_t int
#else
# include <sys/types.h>
# include <sys/time.h>
# include <sys/ioctl.h>
# include <sys/socket.h>
# include <sys/sysinfo.h>
# if __def_mzproxy_use_telnetd__ != (0L)
#  include <sys/wait.h>
# endif
# include <stdio.h>
# if __def_mzproxy_use_telnetd__ == (0L)
#  include <stdlib.h>
# else
#  if !defined(__USE_GNU)
#   define __USE_GNU (1)
#  endif
#  if !defined(__USE_XOPEN)
#   define __USE_XOPEN (1)
#  endif
#  include <stdlib.h>
#  undef __USE_XOPEN
#  undef __USE_GNU
# endif
# include <stdarg.h>
# include <string.h>
# include <fcntl.h>
# include <unistd.h>
# include <signal.h>
# include <errno.h>
# include <netdb.h>
# if __def_mzproxy_use_ifaddrs__ != (0L)
#  include <ifaddrs.h>
# endif
# include <netinet/in.h>
# include <net/if.h>
# include <arpa/inet.h>
# if __def_mzproxy_use_telnetd__ != (0L)
#  include <arpa/telnet.h>
# endif
# define mzproxy_socket_t int
# define mzproxy_socklen_t socklen_t
#endif

#if __def_mzproxy_use_telnetd__ != (0L)
# include <termios.h>
#endif

#if __def_mzproxy_use_syslog__ != (0L)
# include <syslog.h>
#endif

/* ---- */

#define __def_mzproxy_minimum_fifo_size__                            (256) /* for small memory: 256 bytes */
#if __def_mzproxy_use_socks__ != (0L)
# if __def_mzproxy_minimum_fifo_size__ < (514)
#  undef __def_mzproxy_minimum_fifo_size__
#  define __def_mzproxy_minimum_fifo_size__                          (514)
# endif
#endif
#define __def_mzproxy_recommanded_fifo_size__                        (1500) /* for normal case: <MTU> bytes (recommanded) */
#define __def_mzproxy_fastest_fifo_size__                            (1500<<1) /* for fast speed: <MTU>*2 bytes */

#define __def_mzproxy_default_fifo_size__                            __def_mzproxy_recommanded_fifo_size__

#define __def_mzproxy_default_connect_address__                      "localhost"

/* ---- */

/* debug options */
#if !defined(__def_mzproxy_rx_dump__)
# define __def_mzproxy_rx_dump__                                     (0L) /* 0=disable, 1=enable */
#endif
#if !defined(__def_mzproxy_tx_dump__)
# define __def_mzproxy_tx_dump__                                     (1L) /* 0=disable, 1=enable */
#endif

/* ---- */

/* session flag */
#define def_mzproxy_session_flag_none                                (0)

#define def_mzproxy_session_flag_rx_client_event                     (1 << 1)
#define def_mzproxy_session_flag_tx_client_event                     (1 << 2)
#define def_mzproxy_session_flag_rx_server_event                     (1 << 3)
#define def_mzproxy_session_flag_tx_server_event                     (1 << 4)
#define def_mzproxy_session_flag_nonblock                            (1 << 5)
#define def_mzproxy_session_flag_client_event                        (def_mzproxy_session_flag_rx_client_event|def_mzproxy_session_flag_tx_client_event)
#define def_mzproxy_session_flag_server_event                        (def_mzproxy_session_flag_rx_server_event|def_mzproxy_session_flag_tx_server_event)
#define def_mzproxy_session_flag_event                               (def_mzproxy_session_flag_client_event|def_mzproxy_session_flag_server_event)

#define def_mzproxy_session_flag_connecting                          (1 << 8)
#define def_mzproxy_session_flag_connected                           (1 << 9)
#define def_mzproxy_session_flag_disconnected_client                 (1 << 10)
#define def_mzproxy_session_flag_disconnected_server                 (1 << 11)

#define def_mzproxy_session_flag_active_connect                      (1 << 16)
#define def_mzproxy_session_flag_active_server                       (1 << 17)

/* ---- */ /* ---- */ /* ---- */ /* ---- */ /* ---- */ /* ---- */ /* ---- */ /* ---- */

#if !defined(__def_mzproxy_use_pragma_pack__)
# define __def_mzproxy_use_pragma_pack__                             (0L)
#endif

#if __def_mzproxy_use_pragma_pack__ != (0L)
# pragma pack(push,8)
#endif

/* ---- */

typedef union mzproxy_const_ptr_with_ptr_tu {
    /* const ptr */
    const void *const_ptr;
    /* ptr */
    void *ptr;
}__mzproxy_const_ptr_with_ptr_t;
#define mzproxy_const_ptr_with_ptr_t                                 __mzproxy_const_ptr_with_ptr_t

/* ---- */

typedef struct mzproxy_shift_fifo_ts {
    size_t size;
    void *alloc_data;
    unsigned char *data;
    
    size_t entry;
}__mzproxy_shift_fifo_t;
#define mzproxy_shift_fifo_t                                         __mzproxy_shift_fifo_t

#define __mzproxy_fifo_t                                             __mzproxy_shift_fifo_t
#define mzproxy_fifo_t                                               __mzproxy_fifo_t

/* ---- */

#if __def_mzproxy_use_telnetd__ != (0L)
/* telnet session info */
typedef struct mzproxy_telnetd_info_ts {
    pid_t pid;
}__mzproxy_telnetd_info_t;
#define mzproxy_telnetd_info_t                                       __mzproxy_telnetd_info_t
#endif

#if __def_mzproxy_use_socks__ != (0L)
# define def_mzproxy_socks_flag_none                                 (0)
# define def_mzproxy_socks_flag_select_method                        (1 << 0)
# define def_mzproxy_socks_flag_select_method_ack                    (1 << 1)
# define def_mzproxy_socks_flag_method                               (1 << 2)
# define def_mzproxy_socks_flag_method_ack                           (1 << 3)

typedef struct mzproxy_socks_info_ts {
    int flags;
    unsigned char select_method[1 /* version */ + 1 /* methods */ + 255 /* method[] */];
    int done_methods;
}__mzproxy_socks_info_t;
#define mzproxy_socks_info_t                                         __mzproxy_socks_info_t
#endif


/* ---- */

/* proxy options */
typedef struct mzproxy_options_ts {
    const char *application_name;
    int do_help;
    int verbose;
#if def_mzproxy_windows == (0L)
    int daemonize;
#endif
    int listen_port;
    long unsigned int active_timeout;
    long unsigned int connect_timeout;
    const char *connect_address;
    int connect_port;
    long unsigned int data_timeout;
#if __def_mzproxy_use_ifaddrs__ != (0L)
    const char *interface_name;
#endif 
    size_t fifo_size;
    unsigned int limit_session_count;
#if __def_mzproxy_use_inet6__ != (0L)
    int use_inet6;
#endif 
    int use_xinetd;
#if defined(mzproxy_telnetd_info_t)
    int use_telnetd;
    const char *issue_file;
    const char *login_program;
#endif
#if defined(mzproxy_socks_info_t)
    int use_socks;
#endif
    int *stop_value_ptr;
}__mzproxy_options_t;
#define mzproxy_options_t                                            __mzproxy_options_t

#define __def_mzproxy_fifo_count__                                   (2)
/* proxy session */
typedef struct mzproxy_session_ts {
    struct mzproxy_session_ts *prev, *next;
    long unsigned int connect_time_stamp, data_time_stamp;
    int flags;
    mzproxy_socket_t accept_socket, client_socket;
    char *connect_address;
    int connect_port;
    mzproxy_fifo_t fifo[ __def_mzproxy_fifo_count__ ];

    /* extension */
#if defined(mzproxy_telnetd_info_t)
    mzproxy_telnetd_info_t *telnetd_info;
#endif
#if defined(mzproxy_socks_info_t)
    mzproxy_socks_info_t *socks_info;
#endif
}__mzproxy_session_t;
#define mzproxy_session_t                                            __mzproxy_session_t

/* main control */
typedef struct mzproxy_control_ts {
    int *stop_service;
    mzproxy_options_t *options;
#if defined(mzproxy_telnetd_info_t)
    char *login_program_argv[2];
#endif
    mzproxy_socket_t socket;
    mzproxy_session_t *session;
    unsigned int session_count;
    long unsigned int time_stamp;
    int max_fd;
    fd_set fd_set_rx, fd_set_tx;
    int events;
}__mzproxy_control_t;
#define mzproxy_control_t                                            __mzproxy_control_t

/* ---- */

#if __def_mzproxy_use_pragma_pack__ != (0L)
# pragma pack(pop)
#endif

/* ---- */ /* ---- */ /* ---- */ /* ---- */ /* ---- */ /* ---- */ /* ---- */ /* ---- */


#if !defined(__def_mzproxy_source_mzproxy_c__)
def_mzproxy_import_core mzproxy_options_t *mzproxy_init_options(mzproxy_options_t *s_options);
def_mzproxy_import_core int mzproxy_main_process(mzproxy_options_t *s_options);
#endif

/* ---- */

#define def_mzproxy_log_level_emergency                              (0) /* LOG_EMERG   : system is unusable */
#define def_mzproxy_log_level_alert                                  (1) /* LOG_ALERT   : action must be taken immediately */
#define def_mzproxy_log_level_critical                               (2) /* LOG_CRIT    : critical conditions */
#define def_mzproxy_log_level_error                                  (3) /* LOG_ERR     : error conditions */
#define def_mzproxy_log_level_warning                                (4) /* LOG_WARNING : warning conditions */
#define def_mzproxy_log_level_notice                                 (5) /* LOG_NOTICE  : normal, but significant, condition */
#define def_mzproxy_log_level_info                                   (6) /* LOG_INFO    : informational message */
#define def_mzproxy_log_level_debug                                  (7) /* LOG_DEBUG   : debug-level message */

#define def_mzproxy_log_flag_none                                    (0)
#define def_mzproxy_log_flag_get                                     (1 << 4)
#define def_mzproxy_log_flag_console                                 (1 << 5)
#define def_mzproxy_log_flag_syslog                                  (1 << 6)

#define def_mzproxy_log_level_default                                def_mzproxy_log_level_info
#define def_mzproxy_log_level_mask                                   (0x0000000f)
#define def_mzproxy_log_flags_mask                                   (0xfffffff0)

#if !defined(__def_mzproxy_source_mzproxy_log_c__)
def_mzproxy_import_core void mzproxy_use_log(void);
def_mzproxy_import_core void mzproxy_unuse_log(void);

def_mzproxy_import_core int mzproxy_log_level(int s_log_level);

def_mzproxy_import_core void mzproxy_log_puts(int s_log_level, const char *s_log_message);
def_mzproxy_import_core void mzproxy_log_vprintf(int s_log_level, const char *s_format, va_list s_var);
def_mzproxy_import_core void mzproxy_log_printf(int s_log_level, const char *s_format, ...) mzproxy_printf_varg_check(2,3);
def_mzproxy_import_core void mzproxy_log_dump(int s_log_level, const void *s_data, size_t s_size);
#endif

/* ---- */

#if !defined(__def_mzproxy_source_mzproxy_time_c__)
def_mzproxy_import_core long unsigned int mzproxy_time_stamp_sec(void);
#endif

/* ---- */

#if !defined(__def_mzproxy_source_mzproxy_socket_c__)
def_mzproxy_import_core int __mzproxy_init_network(int s_bool);
def_mzproxy_import_core void mzproxy_closesocket(mzproxy_socket_t s_socket);
def_mzproxy_import_core const char * mzproxy_inet_ntop(int s_family, const void *s_inX_addr, char *s_address, mzproxy_socklen_t s_address_length);
def_mzproxy_import_core int mzproxy_nonblock_socket(mzproxy_socket_t s_socket, int s_enable);
def_mzproxy_import_core int mzproxy_linger_on(mzproxy_socket_t s_socket);
def_mzproxy_import_core int mzproxy_reuse_socket(mzproxy_socket_t s_socket);
def_mzproxy_import_core int mzproxy_oobinline_socket(mzproxy_socket_t s_socket);
def_mzproxy_import_core int mzproxy_tos_lowdelay_socket(mzproxy_socket_t s_socket);
def_mzproxy_import_core ssize_t mzproxy_send(mzproxy_socket_t s_socket, const void *s_data, size_t s_size);
def_mzproxy_import_core ssize_t mzproxy_recv(mzproxy_socket_t s_socket, void *s_data, size_t s_size);
#endif
#define mzproxy_use_network()                                        __mzproxy_init_network(1)
#define mzproxy_unuse_network()                                      __mzproxy_init_network(0)

/* ---- */

#if !defined(__def_mzproxy_source_mzproxy_fifo_c__)
def_mzproxy_import_core mzproxy_shift_fifo_t *mzproxy_new_shift_fifo(mzproxy_shift_fifo_t *s_fifo, void *s_data, size_t s_size);
def_mzproxy_import_core void mzproxy_delete_shift_fifo(mzproxy_shift_fifo_t *s_fifo);
def_mzproxy_import_core size_t mzproxy_get_size_shift_fifo(mzproxy_shift_fifo_t *s_fifo);
def_mzproxy_import_core size_t mzproxy_set_size_shift_fifo(mzproxy_shift_fifo_t *s_fifo, size_t s_size);
def_mzproxy_import_core void *mzproxy_peek_shift_fifo(mzproxy_shift_fifo_t *s_fifo, size_t s_offset);
def_mzproxy_import_core ssize_t mzproxy_push_shift_fifo(mzproxy_shift_fifo_t *s_fifo, const void *s_data, size_t s_size);
def_mzproxy_import_core ssize_t mzproxy_pop_shift_fifo(mzproxy_shift_fifo_t *s_fifo, void *s_data, size_t s_size);
#endif

#define mzproxy_new_fifo(m_fifo,m_data,m_size)                       mzproxy_new_shift_fifo(m_fifo,m_data,m_size)
#define mzproxy_delete_fifo(m_fifo)                                  mzproxy_delete_shift_fifo(m_fifo)
#define mzproxy_get_size_fifo(m_fifo)                                mzproxy_get_size_shift_fifo(m_fifo)
#define mzproxy_set_size_fifo(m_fifo,m_size)                         mzproxy_set_size_shift_fifo(m_fifo,m_size)
#define mzproxy_peek_fifo(m_fifo,m_offset)                           mzproxy_peek_shift_fifo(m_fifo,m_offset)
#define mzproxy_push_fifo(m_fifo,m_data,m_size)                      mzproxy_push_shift_fifo(m_fifo,m_data,m_size)
#define mzproxy_pop_fifo(m_fifo,m_data,m_size)                       mzproxy_pop_shift_fifo(m_fifo,m_data,m_size)

#define mzproxy_peek_byte_fifo(m_fifo,m_offset)                      (*((unsigned char *)mzproxy_peek_fifo(m_fifo,m_offset)))
#define mzproxy_peek_word_fifo(m_fifo,m_offset)                      (*((unsigned short *)mzproxy_peek_fifo(m_fifo,m_offset)))
#define mzproxy_peek_dword_fifo(m_fifo,m_offset)                     (*((unsigned int *)mzproxy_peek_fifo(m_fifo,m_offset)))

#define mzproxy_poke_byte_fifo(m_fifo,m_offset,m_value)              *((unsigned char *)mzproxy_peek_fifo(m_fifo,m_offset))=(unsigned char)m_value

/* ---- */

#if !defined(__def_mzproxy_source_mzproxy_session_c__)
def_mzproxy_import_core mzproxy_session_t * mzproxy_free_session(mzproxy_control_t *s_control, mzproxy_session_t *s_session, int s_nokill);
def_mzproxy_import_core mzproxy_session_t * mzproxy_new_session(mzproxy_control_t *s_control, mzproxy_socket_t s_accept_socket);
def_mzproxy_import_core void mzproxy_del_session(mzproxy_control_t *s_control, mzproxy_session_t *s_session);
#endif

/* ---- */

#endif

/* vim: set expandtab: */
/* End of source */
