/*
 Copyright (C) JAEHYUK CHO
 All rights reserved.
 
 Author: JAEHYUK CHO <minzkn@minzkn.com>
 
 http://www.minzkn.com
*/

#if !defined(__def_mzproxy_source_main_c__)
# define __def_mzproxy_source_main_c__ "main.c"

#include "mzproxy.h"

static void mzproxy_signal(int s_signal);
static void (mzproxy_get_options)(int s_argc, char **s_argv, mzproxy_options_t *s_options);

int main(int s_argc, char **s_argv);

static int g_mzproxy_stop_value = 0;

static void mzproxy_signal(int s_signal)
{
    g_mzproxy_stop_value = 1;

    /* reinstall signal */
    (void)signal(s_signal, mzproxy_signal);
}

static void (mzproxy_get_options)(int s_argc, char **s_argv, mzproxy_options_t *s_options)
{
    int s_index = 1, s_sub_options, s_few_argument = 0;

    if(mzproxy_builtin_expect(s_argc <= 1, 0)) {
        s_options->do_help = 1;
        return;
    }

    while(s_index < s_argc) {
        s_sub_options = s_argc - (s_index + 1);

        if((strcmp(s_argv[s_index], "--help") == 0) || (strcmp(s_argv[s_index], "-h") == 0))s_options->do_help = 1;
        else if((strcmp(s_argv[s_index], "--verbose") == 0) || (strcmp(s_argv[s_index], "-v") == 0)) {
            s_options->verbose = 1;

            mzproxy_log_level(def_mzproxy_log_level_debug | (mzproxy_log_level(def_mzproxy_log_flag_get) & def_mzproxy_log_flags_mask));
        }
#if def_mzproxy_windows == (0L)
        else if((strcmp(s_argv[s_index], "--daemon") == 0) || (strcmp(s_argv[s_index], "-d") == 0)) {
            s_options->daemonize = 1;
            
            mzproxy_log_level(def_mzproxy_log_flag_syslog | mzproxy_log_level(def_mzproxy_log_flag_get));
        }
#endif  
        else if((strcmp(s_argv[s_index], "--cp") == 0) || (strcmp(s_argv[s_index], "-p") == 0)) {
            if(s_sub_options < 1){ s_few_argument = 1; break; }
            s_options->listen_port = (int)atoi(s_argv[++s_index]);
        }
        else if((strcmp(s_argv[s_index], "--timeout") == 0) || (strcmp(s_argv[s_index], "-t") == 0)) {
            if(s_sub_options < 1){ s_few_argument = 1; break; }
            s_options->connect_timeout = (long unsigned int)atoi(s_argv[++s_index]);
        }
        else if(strcmp(s_argv[s_index], "--dt") == 0) {
            if(s_sub_options < 1){ s_few_argument = 1; break; }
            s_options->data_timeout = (long unsigned int)atoi(s_argv[++s_index]);
        }
        else if((strcmp(s_argv[s_index], "--sa") == 0) || (strcmp(s_argv[s_index], "-A") == 0)) {
            if(s_sub_options < 1){ s_few_argument = 1; break; }
            s_options->connect_address = s_argv[++s_index];
        }
        else if((strcmp(s_argv[s_index], "--sp") == 0) || (strcmp(s_argv[s_index], "-S") == 0)) {
            if(s_sub_options < 1){ s_few_argument = 1; break; }
            s_options->connect_port = (int)atoi(s_argv[++s_index]);
        }
#if __def_mzproxy_use_ifaddrs__ != (0L)
        else if(strcmp(s_argv[s_index], "--interface") == 0) {
            if(s_sub_options < 1){ s_few_argument = 1; break; }
            s_options->interface_name = s_argv[++s_index];
        }
#endif  
        else if((strcmp(s_argv[s_index], "--buffer") == 0) || (strcmp(s_argv[s_index], "-b") == 0)) {
            if(s_sub_options < 1){ s_few_argument = 1; break; }
            s_options->fifo_size = (size_t)atoi(s_argv[++s_index]);
        }
        else if((strcmp(s_argv[s_index], "--session") == 0) || (strcmp(s_argv[s_index], "-s") == 0)) {
            if(s_sub_options < 1){ s_few_argument = 1; break; }
            s_options->limit_session_count = (unsigned int)atoi(s_argv[++s_index]);
        }
#if __def_mzproxy_use_inet6__ != (0L)
        else if((strcmp(s_argv[s_index], "-6") == 0) || (strcmp(s_argv[s_index], "--inet6") == 0))s_options->use_inet6 = 1;
#endif       
        else if((strcmp(s_argv[s_index], "--xinetd") == 0) || (strcmp(s_argv[s_index], "-x") == 0))s_options->use_xinetd = 1;
#if defined(mzproxy_telnetd_info_t)
        else if(strcmp(s_argv[s_index], "--telnetd") == 0)s_options->use_telnetd = 1;
        else if(strcmp(s_argv[s_index], "--issue") == 0) {
            if(s_sub_options < 1){ s_few_argument = 1; break; }
            s_options->issue_file = s_argv[++s_index];
        }
        else if(strcmp(s_argv[s_index], "--login") == 0) {
            if(s_sub_options < 1){ s_few_argument = 1; break; }
            s_options->login_program = s_argv[++s_index];
        }
#endif
#if __def_mzproxy_use_socks__ != (0L)
        else if(strcmp(s_argv[s_index], "--socks") == 0)s_options->use_socks = 1;
#endif
        else s_options->do_help = 1;
        
        s_index++;
    }

    if(mzproxy_builtin_expect(s_few_argument != 0, 0)) {
        if(s_options->do_help == 0) {
            s_options->do_help = 1;
            mzproxy_log_printf(def_mzproxy_log_level_error, "few argument \"%s\"", s_argv[s_index]);
        }
    }

    /* check options */
    if(mzproxy_builtin_expect(s_options->fifo_size < ((size_t)__def_mzproxy_minimum_fifo_size__), 0))s_options->fifo_size = (size_t)__def_mzproxy_minimum_fifo_size__;

    if(s_options->verbose != 0) {
#if __def_mzproxy_use_inet6__ != (0L)
        mzproxy_log_printf(def_mzproxy_log_level_info, "verbose mode (inet%d,%s)", (s_options->use_inet6 == 0) ? 4 : 6, (s_options->use_xinetd == 0) ? "standalong" : "xinetd");
#else
        mzproxy_log_printf(def_mzproxy_log_level_info, "verbose mode (%s)", (s_options->use_xinetd == 0) ? "standalong" : "xinetd");
#endif 
    }

    if(mzproxy_builtin_expect((s_options->use_xinetd == 0) && (s_options->listen_port <= 0), 0)) {
        if(s_options->do_help == 0) {
            s_options->do_help = 1;
            mzproxy_log_printf(def_mzproxy_log_level_error, "need listen port option ! (\"--cp <listen port>\")");
        }
    }
#if defined(mzproxy_telnetd_info_t)
    if(s_options->use_telnetd == 0)
#endif
    {
        if(mzproxy_builtin_expect((s_options->connect_port <= 0) && (s_options->use_socks == 0), 0)) {
            if(s_options->do_help == 0) {
                s_options->do_help = 1;
                mzproxy_log_printf(def_mzproxy_log_level_error, "need connect port option ! (\"--sp <connect port>\")");
            }
        }

        if(mzproxy_builtin_expect(s_options->connect_address == ((const char *)0), 0)) {
            static const char cg_default_connect_address[] = { __def_mzproxy_default_connect_address__ };
            s_options->connect_address = (const char *)(&cg_default_connect_address[0]);
        }
    }
}

int main(int s_argc, char **s_argv)
{
    mzproxy_options_t s_local_options, *s_options;
    int s_result;
    
    mzproxy_use_log();
    mzproxy_log_level(def_mzproxy_log_flag_console | def_mzproxy_log_level_notice);

    /* get application name */
    s_options = mzproxy_init_options((mzproxy_options_t *)(&s_local_options));
    if((s_argc >= 1) && (s_argv[0] != ((char *)0))) {
        s_options->application_name = strrchr(s_argv[0], '/');
        if(s_options->application_name != ((const char *)0))s_options->application_name = (const char *)(&s_options->application_name[1]);
        else s_options->application_name = s_argv[0];
    }
 
    /* get options */
    mzproxy_get_options(s_argc, s_argv, s_options);
 
    /* print help */
    if(mzproxy_builtin_expect(s_options->do_help != 0, 0)) {
        mzproxy_log_level(def_mzproxy_log_flag_console | def_mzproxy_log_level_notice);
        mzproxy_log_printf(def_mzproxy_log_level_notice,
            "mzproxy v2.0.0 build 0 (%s %s)\n"
            "Copyrights (C) JaeHyuk Cho <minzkn@minzkn.com> - All rights reserved.\n\n"
            "usage: %s [<options>]\noptions:\n"
            "\t-v, --verbose               : verbose mode (packet dump)\n"
#if def_mzproxy_windows == (0L)
            "\t-d, --daemon                : daemonize process\n"
#endif
            "\t-p, --cp=<listen port>      : listen for connections on port\n"
            "\t-A, --sa=<connect address>  : redirect address (default: \"%s\")\n"
            "\t-S, --sp=<connect port>     : redirect port\n"
            "\t    --at=<connect timeout>  : active timeout (default: %lu sec)\n"
            "\t-t, --timeout=<connect timeout> : connect timeout (default: %lu sec)\n"
            "\t    --dt=<data timeout>     : data timeout (default: %lu sec)\n"
#if __def_mzproxy_use_ifaddrs__ != (0L)
            "\t    --interface=<interface> : specify the network interface to listen on\n"
#endif   
            "\t-b, --buffer=<buffer size>  : rx/tx fixed buffer size (default: %lu bytes x 2)\n"
            "\t-s, --session=<limit count> : session limit count (default: %u)\n"
#if __def_mzproxy_use_inet6__ != (0L) 
            "\t-6, --inet6                 : use IPv6 (default: %d)\n"
#endif   
            "\t-x, --xinetd                : use xinetd mode (default: %d)\n"
#if defined(mzproxy_telnetd_info_t)
            "\t    --telnetd               : use telnetd mode (default: %d)\n"
            "\t    --issue                 : issue file (default: %s)\n"
            "\t    --login                 : login program (default: %s)\n"
#endif
#if __def_mzproxy_use_socks__ != (0L)
            "\t    --socks                 : use socks protocol (default: %d)\n"
#endif            
            "%s",
            __DATE__, __TIME__,
            s_options->application_name,
            (s_options->connect_address == ((const char *)0)) ? __def_mzproxy_default_connect_address__ : s_options->connect_address,
            s_options->active_timeout,
            s_options->connect_timeout,
            s_options->data_timeout,
            (unsigned long)__def_mzproxy_default_fifo_size__,
            s_options->limit_session_count,
#if __def_mzproxy_use_inet6__ != (0L)
            s_options->use_inet6,
#endif   
            s_options->use_xinetd,
#if defined(mzproxy_telnetd_info_t)
            s_options->use_telnetd,
            s_options->issue_file,
            s_options->login_program,
#endif
#if __def_mzproxy_use_socks__ != (0L)
            s_options->use_socks,
#endif
            "\n"
        );
    
        mzproxy_unuse_log();
        
        return(EXIT_SUCCESS);
    }
    
    mzproxy_log_puts(def_mzproxy_log_level_info, "Starting. . .");
            
    /* install signal */
#if defined(SIGPIPE)   
    (void)signal(SIGPIPE, mzproxy_signal);
#endif
#if defined(SIGINT)
    (void)signal(SIGINT, mzproxy_signal);
#endif
#if defined(SIGQUIT)
    (void)signal(SIGQUIT, mzproxy_signal);
#endif
#if defined(SIGTERM)
    (void)signal(SIGTERM, mzproxy_signal);
#endif
#if defined(SIGHUP)
    (void)signal(SIGHUP, mzproxy_signal);
#endif
    s_options->stop_value_ptr = (int *)(&g_mzproxy_stop_value);

    mzproxy_use_network();

    /* do work */
    s_result = mzproxy_main_process(s_options);

    mzproxy_unuse_network();
    
    mzproxy_log_puts(def_mzproxy_log_level_info, "End of mzproxy.");
   
    mzproxy_unuse_log();

    return(s_result);
}

#endif

/* vim: set expandtab: */
/* End of source */
