/*
 Copyright (C) JAEHYUK CHO
 All rights reserved.
 
 Author: JAEHYUK CHO <minzkn@minzkn.com>
 
 http://www.minzkn.com
*/

#if !defined(__def_mzproxy_source_mzproxy_c__)
# define __def_mzproxy_source_mzproxy_c__ "mzproxy.c"

#if !defined(_ISOC99_SOURCE)
# define _ISOC99_SOURCE (1L)
#endif

#if !defined(_GNU_SOURCE)
# define _GNU_SOURCE (1L)
#endif

#include "mzproxy.h"

static void mzproxy_register_event(mzproxy_control_t *s_control);
static void mzproxy_check_listen(mzproxy_control_t *s_control);
static int mzproxy_check_connect(mzproxy_control_t *s_control, mzproxy_session_t *s_session, const char *s_comment);
static int mzproxy_to_server(mzproxy_control_t *s_control, mzproxy_session_t *s_session);
static int mzproxy_to_client(mzproxy_control_t *s_control, mzproxy_session_t *s_session);
static int mzproxy_from_client(mzproxy_control_t *s_control, mzproxy_session_t *s_session);
static int mzproxy_from_server(mzproxy_control_t *s_control, mzproxy_session_t *s_session);
static int mzproxy_do(mzproxy_control_t *s_control);

def_mzproxy_export mzproxy_options_t *mzproxy_init_options(mzproxy_options_t *s_options);
def_mzproxy_export int mzproxy_main_process(mzproxy_options_t *s_options);

/* register event */
static void mzproxy_register_event(mzproxy_control_t *s_control)
{
    mzproxy_session_t *s_session = s_control->session, *s_next_session;
            
    const char *s_connect_address;
 
    FD_ZERO(&s_control->fd_set_rx);
    FD_ZERO(&s_control->fd_set_tx);
 
    s_control->max_fd = 0;
 
    if(s_control->socket != ((mzproxy_socket_t)(-1))) {
        FD_SET(s_control->socket, &s_control->fd_set_rx);
        if(((int)s_control->socket) > s_control->max_fd)s_control->max_fd = (int)s_control->socket;
    }

    while(s_session != ((mzproxy_session_t *)0)) {
        s_next_session = s_session->next;

        s_session->flags &= (~def_mzproxy_session_flag_event);

        if((s_session->flags & (def_mzproxy_session_flag_connecting | def_mzproxy_session_flag_connected)) == def_mzproxy_session_flag_connecting) { /* check connection timeout */
            if(mzproxy_builtin_expect((s_control->time_stamp - s_session->connect_time_stamp) >= s_control->options->connect_timeout, 0)) {
                s_next_session = s_session->next;
                mzproxy_log_printf(def_mzproxy_log_level_error, "connect timeout (id=%d, timeout=%lu/%lu)", (int)s_session->accept_socket, s_control->time_stamp - s_session->connect_time_stamp, s_control->options->connect_timeout);
                mzproxy_del_session(s_control, s_session);
                s_session = s_next_session;

                continue;
            }

            mzproxy_log_printf(def_mzproxy_log_level_info, "check connect timeout (id=%d, timeout=%lu/%lu)", (int)s_session->accept_socket, s_control->time_stamp - s_session->connect_time_stamp, s_control->options->connect_timeout);
        }
        
        if((s_session->flags & def_mzproxy_session_flag_active_server) == def_mzproxy_session_flag_none) { /* check active timeout */
            if(mzproxy_builtin_expect((s_control->time_stamp - s_session->data_time_stamp) >= s_control->options->active_timeout, 0)) {
                s_next_session = s_session->next;
                mzproxy_log_printf(def_mzproxy_log_level_error, "active timeout (id=%d, timeout=%lu/%lu)", (int)s_session->accept_socket, s_control->time_stamp - s_session->data_time_stamp, s_control->options->active_timeout);
                mzproxy_del_session(s_control, s_session);
                s_session = s_next_session;

                continue;
            }

            mzproxy_log_printf(def_mzproxy_log_level_info, "check active timeout (id=%d, timeout=%lu/%lu)", (int)s_session->accept_socket, s_control->time_stamp - s_session->data_time_stamp, s_control->options->active_timeout);
        }

        if(s_control->options->data_timeout > 0ul) { /* check data timeout */
            if(mzproxy_builtin_expect((s_control->time_stamp - s_session->data_time_stamp) >= s_control->options->data_timeout, 0)) {
                s_next_session = s_session->next;
                mzproxy_log_printf(def_mzproxy_log_level_error, "data timeout (id=%d, timeout=%lu/%lu)", (int)s_session->accept_socket, s_control->time_stamp - s_session->data_time_stamp, s_control->options->data_timeout);
                mzproxy_del_session(s_control, s_session);
                s_session = s_next_session;

                continue;
            }

            mzproxy_log_printf(def_mzproxy_log_level_info, "check data timeout (id=%d, timeout=%lu/%lu)", (int)s_session->accept_socket, s_control->time_stamp - s_session->data_time_stamp, s_control->options->data_timeout);
        }

        s_connect_address = (s_session->connect_address != ((char *)0)) ? ((const char *)s_session->connect_address) : s_control->options->connect_address;

        if((s_session->client_socket != ((mzproxy_socket_t)(-1))) &&
           (s_connect_address != ((const char *)0)) &&
           (s_session->connect_port != 0) &&
           ((s_session->flags & def_mzproxy_session_flag_active_connect) == def_mzproxy_session_flag_active_connect) &&
           ((s_session->flags & (def_mzproxy_session_flag_connecting | def_mzproxy_session_flag_connected | def_mzproxy_session_flag_disconnected_server | def_mzproxy_session_flag_disconnected_client)) == def_mzproxy_session_flag_none)) { /* connecting session */
            struct addrinfo s_hints;
            struct addrinfo *s_addrinfo_result;
            struct addrinfo *s_addrinfo;
            int s_addrinfo_check;
            mzproxy_socket_t s_try_socket;

            (void)memset((void *)(&s_hints), 0, sizeof(s_hints));
            s_hints.ai_family = AF_UNSPEC; /* IPv4 or IPv6 */
            s_hints.ai_socktype = SOCK_STREAM; /* UDP or TCP */
            s_hints.ai_flags = AI_PASSIVE; /* wildcard IP address */
            s_hints.ai_protocol = IPPROTO_TCP; /* any protocol */
            s_hints.ai_canonname = NULL;
            s_hints.ai_addr = NULL;
            s_hints.ai_next = NULL;

            s_addrinfo_check = getaddrinfo(s_connect_address, (const char *)0, (const struct addrinfo *)(&s_hints), (struct addrinfo **)(&s_addrinfo_result));
            if(mzproxy_builtin_expect(s_addrinfo_check != 0, 0)) {
                mzproxy_log_printf(def_mzproxy_log_level_error, "getaddrinfo error: %s", (const char *)gai_strerror(s_addrinfo_check));
                mzproxy_del_session(s_control, s_session);
                s_session = s_next_session;

                continue;
            }
            
            s_session->flags |= def_mzproxy_session_flag_connecting;
            s_try_socket = (mzproxy_socket_t)(-1);

            s_addrinfo = s_addrinfo_result;
            while(s_addrinfo != ((struct addrinfo *)0)) {
#if __def_mzproxy_use_inet6__ != (0L)
                if(s_addrinfo->ai_family == AF_INET6) { /* try IPv6 connection */
                    struct sockaddr_in6 *s_sockaddr_in6;
                    
                    s_try_socket = (mzproxy_socket_t)socket(s_addrinfo->ai_family, s_addrinfo->ai_socktype, s_addrinfo->ai_protocol);
                    if(s_try_socket == ((mzproxy_socket_t)(-1))) {
                        continue; 
                    }

                    if(mzproxy_builtin_expect(mzproxy_nonblock_socket(s_try_socket, 1) == (-1), 0)) {
                        mzproxy_log_printf(def_mzproxy_log_level_error, "non-block socket failed ! (id=%d, try_socket=%d)", (int)s_session->accept_socket, (int)s_try_socket);
                        mzproxy_closesocket(s_try_socket);
                        s_try_socket = (mzproxy_socket_t)(-1);
                    
                        continue;
                    }
                    
                    mzproxy_closesocket(s_session->client_socket);
                    s_session->client_socket = s_try_socket;
                    
                    s_sockaddr_in6 = (struct sockaddr_in6 *)s_addrinfo->ai_addr;
                    s_sockaddr_in6->sin6_port = htons(s_session->connect_port);

                    mzproxy_log_printf(def_mzproxy_log_level_info, "connecting(IPv6)... (id=%d, \"%s:%d\")", (int)s_session->accept_socket, s_connect_address, s_session->connect_port);

                    if(connect(s_session->client_socket, (struct sockaddr *)s_sockaddr_in6, (mzproxy_socklen_t)sizeof(struct sockaddr_in6)) == 0) {
                        mzproxy_log_printf(def_mzproxy_log_level_info, "connected (id=%d,\"%s:%d\"[now])", (int)s_session->accept_socket, s_connect_address, s_session->connect_port);
                        (void)mzproxy_nonblock_socket(s_session->client_socket, 0);
                        s_session->data_time_stamp = s_control->time_stamp;
                        
                        s_session->flags &= ~(def_mzproxy_session_flag_connecting | def_mzproxy_session_flag_nonblock);
                        s_session->flags |= def_mzproxy_session_flag_connected;
                    }

                    break;
                }
#endif		

                if(s_addrinfo->ai_family == AF_INET) { /* try IPv4 connection */
                    struct sockaddr_in *s_sockaddr_in;
                
                    s_try_socket = (mzproxy_socket_t)socket(s_addrinfo->ai_family, s_addrinfo->ai_socktype, s_addrinfo->ai_protocol);
                    if(s_try_socket == ((mzproxy_socket_t)(-1))) {
                        continue; 
                    }

                    if(mzproxy_builtin_expect(mzproxy_nonblock_socket(s_try_socket, 1) == (-1), 0)) {
                        mzproxy_log_printf(def_mzproxy_log_level_error, "non-block socket failed ! (id=%d, try_socket=%d)", (int)s_session->accept_socket, (int)s_try_socket);
                        mzproxy_closesocket(s_try_socket);
                        s_try_socket = (mzproxy_socket_t)(-1);
                    
                        continue;
                    }

                    
                    mzproxy_closesocket(s_session->client_socket);
                    s_session->client_socket = s_try_socket;
                    
                    s_sockaddr_in = (struct sockaddr_in *)s_addrinfo->ai_addr;
                    s_sockaddr_in->sin_port = htons(s_session->connect_port);

                    mzproxy_log_printf(def_mzproxy_log_level_info, "connecting(IPv4)... (id=%d, \"%s:%d\")", (int)s_session->accept_socket, s_connect_address, s_session->connect_port);

                    if(connect(s_session->client_socket, (struct sockaddr *)s_sockaddr_in, (mzproxy_socklen_t)sizeof(struct sockaddr_in)) == 0) {
                        mzproxy_log_printf(def_mzproxy_log_level_info, "connected (id=%d,\"%s:%d\"[now])", (int)s_session->accept_socket, s_connect_address, s_session->connect_port);
                        (void)mzproxy_nonblock_socket(s_session->client_socket, 0);
                        s_session->data_time_stamp = s_control->time_stamp;
                        s_session->flags &= ~(def_mzproxy_session_flag_connecting | def_mzproxy_session_flag_nonblock);
                        s_session->flags |= def_mzproxy_session_flag_connected;
                    }
                    
                    break;
                }
                
                s_addrinfo = s_addrinfo->ai_next;
            }

            freeaddrinfo(s_addrinfo_result);

            if(mzproxy_builtin_expect(s_try_socket == ((mzproxy_socket_t)(-1)), 0)) {
                mzproxy_log_printf(def_mzproxy_log_level_error, "oops! where to connection?");
                mzproxy_del_session(s_control, s_session);
                s_session = s_next_session;

                continue;
            }
        }

        if(mzproxy_get_size_fifo((mzproxy_fifo_t *)(&s_session->fifo[1])) > ((size_t)0)) {
            FD_SET(s_session->accept_socket, &s_control->fd_set_tx); /* to client */
            s_session->flags |= def_mzproxy_session_flag_tx_client_event;
            if(((int)s_session->accept_socket) > s_control->max_fd)s_control->max_fd = (int)s_session->accept_socket;
        }

        if(mzproxy_get_size_fifo((mzproxy_fifo_t *)(&s_session->fifo[0])) < s_control->options->fifo_size) {
            FD_SET(s_session->accept_socket, &s_control->fd_set_rx); /* from client */
            s_session->flags |= def_mzproxy_session_flag_rx_client_event;
            if(((int)s_session->accept_socket) > s_control->max_fd)s_control->max_fd = (int)s_session->accept_socket;
        }
        

        if((s_session->flags & (def_mzproxy_session_flag_active_connect | def_mzproxy_session_flag_disconnected_server)) == def_mzproxy_session_flag_active_connect) {
            if((s_session->flags & (def_mzproxy_session_flag_connected | def_mzproxy_session_flag_active_server)) == (def_mzproxy_session_flag_connected | def_mzproxy_session_flag_active_server)) {
                /* tx/rx event */
                if(mzproxy_get_size_fifo((mzproxy_fifo_t *)(&s_session->fifo[0])) > ((size_t)0)) {
                    FD_SET(s_session->client_socket, &s_control->fd_set_tx); /* to server */
                    s_session->flags |= def_mzproxy_session_flag_tx_server_event;
                    if(((int)s_session->client_socket) > s_control->max_fd)s_control->max_fd = (int)s_session->client_socket;
                }
                if(mzproxy_get_size_fifo((mzproxy_fifo_t *)(&s_session->fifo[1])) < s_control->options->fifo_size) {
                    FD_SET(s_session->client_socket, &s_control->fd_set_rx); /* from server */
                    s_session->flags |= def_mzproxy_session_flag_rx_server_event;
                    if(((int)s_session->client_socket) > s_control->max_fd)s_control->max_fd = (int)s_session->client_socket;
                }
            }
            else if((s_session->flags & (def_mzproxy_session_flag_connecting | def_mzproxy_session_flag_connected)) == def_mzproxy_session_flag_connecting) {
                    /* connecting event */
                    FD_SET(s_session->client_socket, &s_control->fd_set_tx); /* to server */
                    FD_SET(s_session->client_socket, &s_control->fd_set_rx); /* from server */
                    s_session->flags |= def_mzproxy_session_flag_server_event;
                    if(((int)s_session->client_socket) > s_control->max_fd)s_control->max_fd = (int)s_session->client_socket;
            }
        }

        s_session = s_next_session;
    }
}

static void mzproxy_check_listen(mzproxy_control_t *s_control)
{
    mzproxy_socket_t s_accept_socket;
    mzproxy_socklen_t s_accept_socket_address_size;
    mzproxy_session_t *s_new_session;
    char s_address_string[ 64 ];
    unsigned int s_address_port;

    /* check on connection (accept) */
    if(mzproxy_builtin_expect(s_control->socket == ((mzproxy_socket_t)(-1)), 0))return;
    if(!FD_ISSET(s_control->socket, &s_control->fd_set_rx))return;
    
    s_control->events--;
#if __def_mzproxy_use_inet6__ != (0L)
    if(s_control->options->use_inet6 != 0) {
        struct sockaddr_in6 s_accept_socket_address6;

        s_accept_socket_address_size = (mzproxy_socklen_t)sizeof(s_accept_socket_address6);
        s_accept_socket = accept(s_control->socket, (struct sockaddr *)(&s_accept_socket_address6), (mzproxy_socklen_t *)(&s_accept_socket_address_size));
        if(mzproxy_builtin_expect(s_accept_socket == (-1), 0)) {
            mzproxy_log_printf(def_mzproxy_log_level_error, "accept faied ! (listen_socket=%d)", (int)s_control->socket);

            return;
        }

        (void)mzproxy_linger_on(s_accept_socket);
        (void)mzproxy_inet_ntop(s_accept_socket_address6.sin6_family, (void *)(&s_accept_socket_address6.sin6_addr), (char *)(&s_address_string[0]), sizeof(s_address_string));
        s_address_port = ntohs(s_accept_socket_address6.sin6_port);
    }
    else
#endif    
    {
        struct sockaddr_in s_accept_socket_address;

        s_accept_socket_address_size = (mzproxy_socklen_t)sizeof(s_accept_socket_address);
        s_accept_socket = accept(s_control->socket, (struct sockaddr *)(&s_accept_socket_address), (mzproxy_socklen_t *)(&s_accept_socket_address_size));
        if(mzproxy_builtin_expect(s_accept_socket == (-1), 0)) {
            mzproxy_log_printf(def_mzproxy_log_level_error, "accept faied ! (listen_socket=%d)", (int)s_control->socket);

            return;
        }

        (void)mzproxy_linger_on(s_accept_socket);
        (void)mzproxy_inet_ntop(AF_INET, (void *)(&s_accept_socket_address.sin_addr), (char *)(&s_address_string[0]), sizeof(s_address_string));
        s_address_port = ntohs(s_accept_socket_address.sin_port);
    }

    if(mzproxy_builtin_expect((s_control->options->limit_session_count > 0u) && (s_control->options->limit_session_count <= s_control->session_count), 0)) { /* limited session count - refuse connection */
        mzproxy_log_printf(def_mzproxy_log_level_warning, "limited session: refused connection (%s:%u)", (char *)(&s_address_string[0]), s_address_port);
        mzproxy_closesocket(s_accept_socket);

        return;
    }

    /* create & add session entry */
    s_new_session = mzproxy_new_session(s_control, s_accept_socket);
    if(s_new_session != ((mzproxy_session_t *)0)) {
        mzproxy_log_printf(def_mzproxy_log_level_info, "connection from %s:%u (id=%d, sessions=%u/%u, redirect_port=%d)", (char *)(&s_address_string[0]), s_address_port, (int)s_new_session->accept_socket, s_control->session_count, s_control->options->limit_session_count, s_new_session->connect_port);
    }
}

static int mzproxy_check_connect(mzproxy_control_t *s_control, mzproxy_session_t *s_session, const char *s_comment)
{
    int s_socket_error;
    mzproxy_socklen_t s_socket_error_length;

    if((s_session->flags & (def_mzproxy_session_flag_connecting | def_mzproxy_session_flag_connected)) == def_mzproxy_session_flag_connected) {
        /* already connected */
        return(1);
    }
    
    if((s_session->flags & (def_mzproxy_session_flag_connecting | def_mzproxy_session_flag_connected)) != def_mzproxy_session_flag_connecting) {
        /* is not connecting */
        return(0);
    }

    s_socket_error_length = (mzproxy_socklen_t)sizeof(s_socket_error);
    if(mzproxy_builtin_expect(getsockopt(s_session->client_socket, SOL_SOCKET, SO_ERROR, (void *)(&s_socket_error), &s_socket_error_length) != 0, 0)) {
        mzproxy_log_printf(def_mzproxy_log_level_error, "can not connect (id=%d,getsockopt)", (int)s_session->accept_socket);
        s_session->flags |= def_mzproxy_session_flag_disconnected_server;

        return(-1);
    }

    if(mzproxy_builtin_expect(s_socket_error != 0, 0)) {
        mzproxy_log_printf(def_mzproxy_log_level_error, "can not connect (id=%d,error)", (int)s_session->accept_socket);
        s_session->flags |= def_mzproxy_session_flag_disconnected_server;

        return(-1);
    }

    /* connected */
    mzproxy_log_printf(def_mzproxy_log_level_info, "connected (id=%d,[%s])", (int)s_session->accept_socket, s_comment);
    (void)mzproxy_nonblock_socket(s_session->client_socket, 0);
    s_session->data_time_stamp = s_control->time_stamp;

    s_session->flags &= ~(def_mzproxy_session_flag_connecting | def_mzproxy_session_flag_nonblock);
    s_session->flags |= def_mzproxy_session_flag_connected;

    /* connected, next time */
    return(0);
}

/* to server: fifo[0] -> server */
static int mzproxy_to_server(mzproxy_control_t *s_control, mzproxy_session_t *s_session)
{
    int s_send_bytes, s_check_connect;

    if((s_session->flags & def_mzproxy_session_flag_tx_server_event) == def_mzproxy_session_flag_none)return(0);
    
    if(!FD_ISSET(s_session->client_socket, &s_control->fd_set_tx))return(0);

    s_control->events--;

    s_check_connect = mzproxy_check_connect(s_control, s_session, "tx");
    if(s_check_connect <= 0)return(s_check_connect);
    
    if((s_session->flags & def_mzproxy_session_flag_active_server) == def_mzproxy_session_flag_none)return(0);
    
    if(mzproxy_get_size_fifo((mzproxy_fifo_t *)(&s_session->fifo[0])) <= ((size_t)0))return(0);

    do {
        size_t s_from = (size_t)0, s_to = (size_t)0;

#if defined(mzproxy_telnetd_info_t)
        if(s_session->telnetd_info != ((mzproxy_telnetd_info_t *)0)) {
            while(s_from < mzproxy_get_size_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]))) {
                if(mzproxy_peek_byte_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), s_from) != IAC) { /* pt data */
                    if(s_to != s_from) {
                        mzproxy_poke_byte_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), s_to, mzproxy_peek_byte_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), s_from));
                    }
                    s_to++, s_from++;
                }
                else {
                    if(mzproxy_builtin_expect((mzproxy_get_size_fifo((mzproxy_fifo_t *)(&s_session->fifo[0])) - s_from) < ((size_t)3), 0))break;

                    switch(mzproxy_peek_byte_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), s_from + ((size_t)1))) { 
                        case DO:
                        case DONT:
                        case WILL:
                        case WONT:
                             s_from += (size_t)3; /* ignore IAC */
                             break;
                        case SB:
                             if(mzproxy_peek_byte_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), s_from + ((size_t)2)) == TELOPT_NAWS) { /* setup pt window size */
                                 struct winsize s_winsize;

                                 if((mzproxy_get_size_fifo((mzproxy_fifo_t *)(&s_session->fifo[0])) - s_from) < ((size_t)9))break;
                                 s_winsize.ws_col = ((unsigned short int)mzproxy_peek_byte_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), s_from + ((size_t)3)) << 8) | ((unsigned short int)mzproxy_peek_byte_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), s_from + ((size_t)4)));
                                 s_winsize.ws_row = ((unsigned short int)mzproxy_peek_byte_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), s_from + ((size_t)5)) << 8) | ((unsigned short int)mzproxy_peek_byte_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), s_from + ((size_t)6)));
                                 (void)ioctl((int)s_session->client_socket, TIOCSWINSZ, (void *)(&s_winsize));
                                 s_from += (size_t)(3 + 4 + 2);

                                 mzproxy_log_printf(def_mzproxy_log_level_info, "winsize=%d,%d", (int)s_winsize.ws_col, (int)s_winsize.ws_row);
                             }
                             break;
                        default: s_from += (size_t)3; /* ignore IAC */ break;
                    }
                }
            }
        }
        else s_to = s_from = mzproxy_get_size_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]));
#else        
        s_to = s_from = mzproxy_get_size_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]));
#endif  

        (void)mzproxy_set_size_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), s_to + (mzproxy_get_size_fifo((mzproxy_fifo_t *)(&s_session->fifo[0])) - s_from));
        
        if(s_to <= ((size_t)0))return(0);

#if defined(mzproxy_telnetd_info_t)
        if(s_session->telnetd_info != ((mzproxy_telnetd_info_t *)0)) {
            s_send_bytes = (int)write(s_session->client_socket, (const void *)mzproxy_peek_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), (size_t)0), s_to);
        }
        else
#endif
        {
            s_send_bytes = (int)mzproxy_send(s_session->client_socket, (const void *)mzproxy_peek_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), (size_t)0), s_to);
        }

        if((s_send_bytes == 0) || ((s_send_bytes == (-1)) && (errno != EINTR))) { /* disconnected by server */
            mzproxy_log_printf(def_mzproxy_log_level_info, "disconnected by server (id=%d, tx=%d)", (int)s_session->accept_socket, (int)s_send_bytes);
            s_session->flags |= def_mzproxy_session_flag_disconnected_server;

            return(-1);
        }

        mzproxy_log_printf(def_mzproxy_log_level_info, "client -> server (id=%d, tx=%d)", (int)s_session->accept_socket, (int)s_send_bytes);
#if __def_mzproxy_tx_dump__ != (0L)
        (void)mzproxy_log_dump(def_mzproxy_log_level_debug, (const void *)mzproxy_peek_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), (size_t)0), s_send_bytes);
#endif    

        (void)mzproxy_pop_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), (void *)0, (size_t)s_send_bytes);
    }while(0);
    
    return(0);
}

/* to client: fifo[1] -> client */
static int mzproxy_to_client(mzproxy_control_t *s_control, mzproxy_session_t *s_session)
{
    size_t s_fifo_size;
    int s_send_bytes;
    
    if((s_session->flags & def_mzproxy_session_flag_tx_client_event) == def_mzproxy_session_flag_none)return(0);

    if(!FD_ISSET(s_session->accept_socket, &s_control->fd_set_tx))return(0);

    s_control->events--;

    s_fifo_size = mzproxy_get_size_fifo((mzproxy_fifo_t *)(&s_session->fifo[1]));
    if(s_fifo_size > ((size_t)0)) { /* normal send */
        s_send_bytes = (int)mzproxy_send(s_session->accept_socket, (const void *)mzproxy_peek_fifo((mzproxy_fifo_t *)(&s_session->fifo[1]), (size_t)0), s_fifo_size);
        if((s_send_bytes == 0) || ((s_send_bytes == (-1)) && (errno != EINTR))) { /* disconnected by client */
            mzproxy_log_printf(def_mzproxy_log_level_info, "disconnected by client (id=%d, tx=%d)", (int)s_session->accept_socket, (int)s_send_bytes);
            s_session->flags |= def_mzproxy_session_flag_disconnected_client;

            return(-1);
        }

        mzproxy_log_printf(def_mzproxy_log_level_info, "%s -> client (id=%d, tx=%d)", (s_session->flags & def_mzproxy_session_flag_active_server) ? "server" : "proxy", (int)s_session->accept_socket, (int)s_send_bytes);
#if __def_mzproxy_tx_dump__ != (0L)
        (void)mzproxy_log_dump(def_mzproxy_log_level_debug, (const void *)mzproxy_peek_fifo((mzproxy_fifo_t *)(&s_session->fifo[1]), (size_t)0), s_send_bytes);
#endif   

        s_fifo_size -= mzproxy_pop_fifo((mzproxy_fifo_t *)(&s_session->fifo[1]), (void *)0, (size_t)s_send_bytes);

#if defined(mzproxy_socks_info_t)
        if(s_session->socks_info != ((mzproxy_socks_info_t *)0)) {
            if((s_session->flags & def_mzproxy_session_flag_active_server) == def_mzproxy_session_flag_none) {
                if((s_session->socks_info->flags & (def_mzproxy_socks_flag_select_method | def_mzproxy_socks_flag_select_method_ack)) == def_mzproxy_socks_flag_select_method) {
                    if(s_fifo_size <= ((size_t)0)) {
                        mzproxy_log_printf(def_mzproxy_log_level_debug, "socks select ack (id=%d)", (int)s_session->accept_socket);

                        s_session->socks_info->flags |= def_mzproxy_socks_flag_select_method_ack;
                    }
                }

                if((s_session->socks_info->flags & (def_mzproxy_socks_flag_method | def_mzproxy_socks_flag_method_ack)) == def_mzproxy_socks_flag_method) {
                    if(s_fifo_size <= ((size_t)0)) {
                        mzproxy_log_printf(def_mzproxy_log_level_debug, "socks method ack (id=%d)", (int)s_session->accept_socket);

                        s_session->socks_info->flags |= def_mzproxy_socks_flag_method_ack;
                    }
                }

                if((s_session->socks_info->flags & (def_mzproxy_socks_flag_select_method_ack | def_mzproxy_socks_flag_method_ack)) == (def_mzproxy_socks_flag_select_method_ack | def_mzproxy_socks_flag_method_ack)) {
                    mzproxy_log_printf(def_mzproxy_log_level_debug, "active to connect (id=%d)", (int)s_session->accept_socket);
                    
                    s_session->flags |= def_mzproxy_session_flag_active_server;
                }
            }
        }
#endif
    }
    
    return(0);
}

/* from client: client -> fifo[0] */
static int mzproxy_from_client(mzproxy_control_t *s_control, mzproxy_session_t *s_session)
{
    size_t s_fifo_size;
    int s_recv_bytes;
    
    if((s_session->flags & def_mzproxy_session_flag_rx_client_event) == def_mzproxy_session_flag_none)return(0);

    if(!FD_ISSET(s_session->accept_socket, &s_control->fd_set_rx))return(0);

    s_control->events--;

    s_fifo_size = mzproxy_get_size_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]));
    if(s_fifo_size < s_control->options->fifo_size) {
        s_recv_bytes = (int)mzproxy_recv(s_session->accept_socket, (void *)mzproxy_peek_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), (size_t)0), s_control->options->fifo_size - s_fifo_size);

        if((s_recv_bytes == 0) || ((s_recv_bytes == (-1)) && (errno != EINTR))) { /* disconnected by client */
            mzproxy_log_printf(def_mzproxy_log_level_info, "disconnected by client (id=%d, rx=%d)", (int)s_session->accept_socket, (int)s_recv_bytes);
            s_session->flags |= def_mzproxy_session_flag_disconnected_client;

            return(-1);
        }

        s_session->data_time_stamp = s_control->time_stamp;

        mzproxy_log_printf(def_mzproxy_log_level_info, "from client (id=%d, rx=%d)", (int)s_session->accept_socket, (int)s_recv_bytes);
#if __def_mzproxy_rx_dump__ != (0L)
        (void)mzproxy_log_dump(def_mzproxy_log_level_debug, (const void *)mzproxy_peek_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), (size_t)0), s_recv_bytes);
#endif   

        mzproxy_push_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), (const void *)0, (size_t)s_recv_bytes);
        s_fifo_size = mzproxy_get_size_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]));
    }

    if((s_session->flags & def_mzproxy_session_flag_active_server) == def_mzproxy_session_flag_none) {
#if defined(mzproxy_socks_info_t)
        if(s_session->socks_info != ((mzproxy_socks_info_t *)0)) {
            if((s_session->socks_info->flags & def_mzproxy_socks_flag_select_method) == def_mzproxy_socks_flag_none) {
                if(s_fifo_size >= ((size_t)2)) {
                    unsigned char s_temp[ 2 ];

                    /* temp[0] = socks version */
                    s_temp[0] = mzproxy_peek_byte_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), (size_t)0);

                    if(s_temp[0] != ((unsigned char)5)) { /* socks version check */
                        mzproxy_log_printf(def_mzproxy_log_level_error, "invalid socks version (id=%d, socks_version=%u)", (int)s_session->accept_socket, (unsigned int)s_temp[0]);

                        return(-1);
                    }
                   
                    /* temp[1] = socks methods */
                    s_temp[1] = mzproxy_peek_byte_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), (size_t)1);
                    if(s_fifo_size >= (((size_t)2) + ((size_t)s_temp[1]))) {
                        mzproxy_pop_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), (void *)(&s_session->socks_info->select_method[0]), (((size_t)2) + ((size_t)s_temp[1])));
                        s_fifo_size = mzproxy_get_size_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]));
                        
                        mzproxy_log_printf(def_mzproxy_log_level_debug,
                            "socks select method (id=%d, socks_version=%u, socks_methods=%u, fifo_size=%lu)",
                            (int)s_session->accept_socket, (unsigned int)s_session->socks_info->select_method[0], (unsigned int)s_session->socks_info->select_method[1], (unsigned long)s_fifo_size);

                        for(s_session->socks_info->done_methods = 0;s_session->socks_info->done_methods < ((int)s_session->socks_info->select_method[1]);s_session->socks_info->done_methods++) {
                            s_temp[1] = s_session->socks_info->select_method[2 + s_session->socks_info->done_methods];
                            mzproxy_push_fifo((mzproxy_fifo_t *)(&s_session->fifo[1]), (void *)(&s_temp[0]), sizeof(s_temp));
                        }
                        s_session->socks_info->done_methods = 0;

                        s_session->socks_info->flags |= def_mzproxy_socks_flag_select_method;
                        if(s_session->socks_info->select_method[1] == ((unsigned char)0u)) {
                            s_session->socks_info->flags |= def_mzproxy_socks_flag_method | def_mzproxy_socks_flag_method_ack;
                        }
                    }
                }
            }
            else if((s_session->socks_info->flags & (def_mzproxy_socks_flag_method | def_mzproxy_socks_flag_select_method_ack)) == def_mzproxy_socks_flag_select_method_ack) {
                if(s_fifo_size >= ((size_t)2)) {
                    unsigned char s_socks_version, s_current_method;

                    s_socks_version = mzproxy_peek_byte_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), (size_t)0);
                    
                    if(s_socks_version != s_session->socks_info->select_method[0]) { /* socks version check */
                        mzproxy_log_printf(def_mzproxy_log_level_error, "invalid socks version (id=%d, socks_version=%u)", (int)s_session->accept_socket, (unsigned int)s_socks_version);

                        return(-1);
                    }

                    s_current_method = s_session->socks_info->select_method[2 + s_session->socks_info->done_methods];

                    if(s_current_method == ((unsigned char)0x00)) { /*  NO AUTHENTICATION REQUIRED */
                        unsigned char s_socks_cmd;
                        
                        s_socks_cmd = mzproxy_peek_byte_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), (size_t)1);
                        if(s_socks_cmd == ((unsigned char)0x01)) { /* CMD: CONNECT */
                            if(s_fifo_size >= ((size_t)(1 + 1 + 1 + 1))) {
                                unsigned char s_socks_rsv, s_socks_atyp;
                                
                                int s_socks_dst_addr_size_valid; 
                                size_t s_socks_dst_addr_size;

                                s_socks_rsv = mzproxy_peek_byte_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), (size_t)2);
                                if(s_socks_rsv != ((unsigned char)0x00)) {
                                    mzproxy_log_printf(def_mzproxy_log_level_error, "invalid RSV != 00H (id=%d, socks_version=%u, current_method=%02XH, cmd=%02XH, rsv=%02XH)", (int)s_session->accept_socket, (unsigned int)s_socks_version, (unsigned int)s_current_method, (unsigned int)s_socks_cmd, (unsigned int)s_socks_rsv);

                                    return(-1);
                                }

                                s_socks_atyp = mzproxy_peek_byte_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), (size_t)3);
                                if(s_socks_atyp == ((unsigned char)0x01)) { /* IPv4 (4 bytes) */
                                    s_socks_dst_addr_size_valid = 1;
                                    s_socks_dst_addr_size = (size_t)4u;
                                }
                                else if(s_socks_atyp == ((unsigned char)0x03)) { /* DOMAINNAME */
                                    s_socks_dst_addr_size_valid = 0;
                                    s_socks_dst_addr_size = (size_t)0u;

                                    while(s_fifo_size >= (((size_t)(1 + 1 + 1 + 1)) + s_socks_dst_addr_size)) {
                                        if(mzproxy_peek_byte_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), ((size_t)(1 + 1 + 1 + 1)) + s_socks_dst_addr_size) == ((unsigned char)0x00)) {
                                            s_socks_dst_addr_size_valid = 1;
                                            ++s_socks_dst_addr_size;
                                            break;
                                        }
                                        ++s_socks_dst_addr_size;
                                    }
                                }
                                else if(s_socks_atyp == ((unsigned char)0x04)) { /* IPv6 (16 bytes) */
                                    s_socks_dst_addr_size_valid = 1;
                                    s_socks_dst_addr_size = (size_t)16u;
                                }
                                else {
                                    s_socks_dst_addr_size_valid = 0;
                                    s_socks_dst_addr_size = (size_t)0u;
                                }
                            
                                if(s_socks_dst_addr_size_valid != 0) {
                                    if(s_fifo_size >= (((size_t)(1 + 1 + 1 + 1)) + s_socks_dst_addr_size + ((size_t)2))) {
                                        unsigned int s_socks_port;
                                    
                                        s_socks_port  = ((unsigned int)mzproxy_peek_byte_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), ((size_t)(1 + 1 + 1 + 1)) + s_socks_dst_addr_size)) << 8;
                                        s_socks_port |= (unsigned int)mzproxy_peek_byte_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), ((size_t)(1 + 1 + 1 + 1)) + s_socks_dst_addr_size + ((size_t)1));

                                        if(s_session->connect_address != ((char *)0)) {
                                            free((void *)s_session->connect_address);
                                            s_session->connect_address = (char *)0;
                                        }

                                        if(s_socks_atyp == ((unsigned char)0x01)) { /* IPv4 (4 bytes) */
                                            char s_socks_ipv4_string[ 16 ];

                                            (void)sprintf((char *)(&s_socks_ipv4_string[0]), "%u.%u.%u.%u", (unsigned int)
                                                ((unsigned int)mzproxy_peek_byte_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), ((size_t)(1 + 1 + 1 + 1)) + ((size_t)0))),
                                                ((unsigned int)mzproxy_peek_byte_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), ((size_t)(1 + 1 + 1 + 1)) + ((size_t)1))),
                                                ((unsigned int)mzproxy_peek_byte_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), ((size_t)(1 + 1 + 1 + 1)) + ((size_t)2))),
                                                ((unsigned int)mzproxy_peek_byte_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), ((size_t)(1 + 1 + 1 + 1)) + ((size_t)3))));

#if def_mzproxy_windows != 0L
                                            s_session->connect_address = _strdup((const char *)(&s_socks_ipv4_string[0]));
#else
                                            s_session->connect_address = strdup((const char *)(&s_socks_ipv4_string[0]));
#endif
                                            if(mzproxy_builtin_expect(s_session->connect_address == ((char *)0), 0)) {
                                                mzproxy_log_printf(def_mzproxy_log_level_error, "not enough memory ! (id=%d, connect_address)", (int)s_session->accept_socket);
                                                
                                                return(-1);
                                            }
                                            s_session->connect_port = s_socks_port;
                                        }
                                        else if(s_socks_atyp == ((unsigned char)0x03)) { /* DOMAINNAME */
#if def_mzproxy_windows != 0L
                                            s_session->connect_address = _strdup((const char *)mzproxy_peek_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), ((size_t)(1 + 1 + 1 + 1))));
#else
                                            s_session->connect_address = strdup((const char *)mzproxy_peek_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), ((size_t)(1 + 1 + 1 + 1))));
#endif					    
                                            if(mzproxy_builtin_expect(s_session->connect_address == ((char *)0), 0)) {
                                                mzproxy_log_printf(def_mzproxy_log_level_error, "not enough memory ! (id=%d, connect_address)", (int)s_session->accept_socket);
                                                
                                                return(-1);
                                            }
                                            s_session->connect_port = s_socks_port;
                                        }
                                        else if(s_socks_atyp == ((unsigned char)0x04)) { /* IPv6 (16 bytes) */
                                            mzproxy_log_printf(def_mzproxy_log_level_error, "not supported socks/\"ATYP: IPv6\" (id=%d, socks_version=%u, current_method=%02XH, cmd=%02XH, atyp=%02XH)", (int)s_session->accept_socket, (unsigned int)s_socks_version, (unsigned int)s_current_method, (unsigned int)s_socks_cmd, (unsigned int)s_socks_atyp);

                                            return(-1);
                                        }
                                        else {
                                            mzproxy_log_printf(def_mzproxy_log_level_error, "not supported socks/\"ATYP\" (id=%d, socks_version=%u, current_method=%02XH, cmd=%02XH, atyp=%02XH)", (int)s_session->accept_socket, (unsigned int)s_socks_version, (unsigned int)s_current_method, (unsigned int)s_socks_cmd, (unsigned int)s_socks_atyp);

                                            return(-1);
                                        }

                                        /* try connect */
                                        s_session->flags |= def_mzproxy_session_flag_active_connect;
                                        
                                        if(1 || (s_session->flags & def_mzproxy_session_flag_connected) == def_mzproxy_session_flag_connected) {
                                            do {
                                                unsigned char s_temp[ 2 + 8 ];
                                                s_temp[0] = s_socks_version;
                                                s_temp[1] = (unsigned int)0x00;
                                                s_temp[2] = (unsigned int)0x00;
                                                s_temp[3] = (unsigned int)0x01;

                                                s_temp[4] = (unsigned int)59;
                                                s_temp[5] = (unsigned int)6;
                                                s_temp[6] = (unsigned int)148;
                                                s_temp[7] = (unsigned int)189;

                                                s_temp[8] = (unsigned int)0x00u;
                                                s_temp[9] = (unsigned int)0x00u;
                                                mzproxy_push_fifo((mzproxy_fifo_t *)(&s_session->fifo[1]), (void *)(&s_temp[0]), sizeof(s_temp));
                                            }while(0);
                                            
                                            mzproxy_log_printf(def_mzproxy_log_level_debug, "socks connect to \"%s:%u\" (id=%d, socks_version=%u, current_method=%02XH, cmd=%02XH, atyp=%02XH)", s_session->connect_address, (unsigned int)s_session->connect_port, (int)s_session->accept_socket, (unsigned int)s_socks_version, (unsigned int)s_current_method, (unsigned int)s_socks_cmd, (unsigned int)s_socks_atyp);
                        
                                            s_fifo_size -= mzproxy_pop_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), (void *)0, (((size_t)(1 + 1 + 1 + 1)) + ((size_t)s_socks_dst_addr_size) + ((size_t)2)));
                                            ++s_session->socks_info->done_methods;
                                        }
                                    }
                                }
                            }
                        }
                        else if(s_socks_cmd == ((unsigned char)0x02)) { /* CMD: BIND */
                            mzproxy_log_printf(def_mzproxy_log_level_error, "not supported socks/\"CMD: BIND\" (id=%d, socks_version=%u, current_method=%02XH, cmd=%02XH)", (int)s_session->accept_socket, (unsigned int)s_socks_version, (unsigned int)s_current_method, (unsigned int)s_socks_cmd);

                            return(-1);
                        }
                        else if(s_socks_cmd == ((unsigned char)0x03)) { /* CMD: UDP ASSOCIATE */
                            mzproxy_log_printf(def_mzproxy_log_level_error, "not supported socks/\"CMD: UDP ASSOCIATE\" (id=%d, socks_version=%u, current_method=%02XH, cmd=%02XH)", (int)s_session->accept_socket, (unsigned int)s_socks_version, (unsigned int)s_current_method, (unsigned int)s_socks_cmd);

                            return(-1);
                        }
                        else {
                            mzproxy_log_printf(def_mzproxy_log_level_error, "not supported socks/\"CMD\" (id=%d, socks_version=%u, current_method=%02XH, cmd=%02XH)", (int)s_session->accept_socket, (unsigned int)s_socks_version, (unsigned int)s_current_method, (unsigned int)s_socks_cmd);

                            return(-1);
                        }
                    }
                    else if(s_current_method == ((unsigned char)0x01)) { /* GSSAPI */
                        mzproxy_log_printf(def_mzproxy_log_level_error, "not supported socks/\"GSSAPI\" (id=%d, socks_version=%u, current_method=%02XH)", (int)s_session->accept_socket, (unsigned int)s_socks_version, (unsigned int)s_current_method);

                        return(-1);
                    }
                    else if(s_current_method == ((unsigned char)0x02)) { /* USERNAME/PASSWORD */
                        unsigned char s_socks_username_len, s_socks_password_len;

                        s_socks_username_len = mzproxy_peek_byte_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), (size_t)1);
                        
                        if(s_fifo_size >= (((size_t)(1 + 1)) + ((size_t)s_socks_username_len) + ((size_t)1))) {
                            char s_socks_username[ 256 ];

                            if(s_socks_username_len > ((unsigned char)0u)) {
                                (void)memcpy((void *)(&s_socks_username[0]), (const void *)mzproxy_peek_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), (size_t)2), (size_t)s_socks_username_len);
                            }
                            s_socks_username[(size_t)s_socks_username_len] = (char)0;
                           
                            s_socks_password_len = mzproxy_peek_byte_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), ((size_t)(1 + 1)) + ((size_t)s_socks_username_len));
                            if(s_fifo_size >= (((size_t)(1 + 1)) + ((size_t)s_socks_username_len) + ((size_t)1) + ((size_t)s_socks_password_len))) {
                                char s_socks_password[ 256 ];
                                unsigned char s_temp[ 2 ];
                            
                                if(s_socks_password_len > ((unsigned char)0u)) {
                                    (void)memcpy((void *)(&s_socks_password[0]), (const void *)mzproxy_peek_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), ((size_t)2) + ((size_t)s_socks_username_len) + ((size_t)1)), (size_t)s_socks_password_len);
                                }
                                s_socks_password[(size_t)s_socks_password_len] = (char)0;

                                s_temp[0] = s_socks_version;
                                s_temp[1] = (unsigned int)0x00u; /* ACK=00H, NACK=!00H */
                                mzproxy_push_fifo((mzproxy_fifo_t *)(&s_session->fifo[1]), (void *)(&s_temp[0]), sizeof(s_temp));

                                mzproxy_log_printf(def_mzproxy_log_level_debug,
                                    "socks method (id=%d, socks_version=%u, method[%d/%u]=%02XH, fifo_size=%lu, username=\"%s\"[%u bytes], password=\"%s\"[%u bytes])",
                                    (int)s_session->accept_socket,
                                    (unsigned int)s_session->socks_info->select_method[0],
                                    s_session->socks_info->done_methods,
                                    (unsigned int)s_session->socks_info->select_method[1],
                                    (unsigned int)s_session->socks_info->select_method[2 + s_session->socks_info->done_methods],
                                    (unsigned long)s_fifo_size,
                                    (char *)(&s_socks_username[0]), (unsigned int)s_socks_username_len,
                                    (char *)(&s_socks_password[0]), (unsigned int)s_socks_password_len);

                                s_fifo_size -= mzproxy_pop_fifo((mzproxy_fifo_t *)(&s_session->fifo[0]), (void *)0, ((size_t)(1 + 1)) + ((size_t)s_socks_username_len) + ((size_t)1) + ((size_t)s_socks_password_len));
                                ++s_session->socks_info->done_methods;
                            }
                        }
                    }
                    else if((s_current_method >= ((unsigned char)0x03)) && (s_current_method <= ((unsigned char)0x7f))) { /* IANA ASSIGNED */
                        mzproxy_log_printf(def_mzproxy_log_level_error, "not supported socks/\"IANA ASSIGNED\" (id=%d, socks_version=%u, current_method=%02XH)", (int)s_session->accept_socket, (unsigned int)s_socks_version, (unsigned int)s_current_method);

                        return(-1);
                    }
                    else if((s_current_method >= ((unsigned char)0x80)) && (s_current_method <= ((unsigned char)0xfe))) { /* RESERVED FOR PRIVATE METHODS */
                        mzproxy_log_printf(def_mzproxy_log_level_error, "not supported socks/\"RESERVED FOR PRIVATE METHODS\" (id=%d, socks_version=%u, current_method=%02XH)", (int)s_session->accept_socket, (unsigned int)s_socks_version, (unsigned int)s_current_method);

                        return(-1);
                    }
                    else { /* NO ACCEPTABLE METHODS */
                        mzproxy_log_printf(def_mzproxy_log_level_error, "not supported socks/\"NO ACCEPTABLE METHODS\" (id=%d, socks_version=%u, current_method=%02XH)", (int)s_session->accept_socket, (unsigned int)s_socks_version, (unsigned int)s_current_method);

                        return(-1);
                    }

                    if(s_session->socks_info->done_methods >= s_session->socks_info->select_method[1]) {
                        s_session->socks_info->flags |= def_mzproxy_socks_flag_method;
                    }
                }
            }
        }
#endif
    }

    return(0);
}

/* from server: server->fifo[1] */
static int mzproxy_from_server(mzproxy_control_t *s_control, mzproxy_session_t *s_session)
{
    int s_recv_bytes, s_check_connect;
    
    if((s_session->flags & def_mzproxy_session_flag_rx_server_event) == def_mzproxy_session_flag_none)return(0);
    
    if(!FD_ISSET(s_session->client_socket, &s_control->fd_set_rx))return(0);
    
    s_control->events--;

    s_check_connect = mzproxy_check_connect(s_control, s_session, "rx");
    if(s_check_connect <= 0)return(s_check_connect);
    
    if((s_session->flags & def_mzproxy_session_flag_active_server) == def_mzproxy_session_flag_none)return(0);

    if(mzproxy_get_size_fifo((mzproxy_fifo_t *)(&s_session->fifo[1])) < s_control->options->fifo_size) {
#if defined(mzproxy_telnetd_info_t)
        if(s_session->telnetd_info != ((mzproxy_telnetd_info_t *)0)) {
            s_recv_bytes = (int)read((int)s_session->client_socket, (void *)mzproxy_peek_fifo((mzproxy_fifo_t *)(&s_session->fifo[1]), (size_t)0), s_control->options->fifo_size - mzproxy_get_size_fifo((mzproxy_fifo_t *)(&s_session->fifo[1])));
        }
        else
#endif
        { 
            s_recv_bytes = (int)mzproxy_recv(s_session->client_socket, (void *)mzproxy_peek_fifo((mzproxy_fifo_t *)(&s_session->fifo[1]), (size_t)0), s_control->options->fifo_size - mzproxy_get_size_fifo((mzproxy_fifo_t *)(&s_session->fifo[1])));
        }

        if((s_recv_bytes == 0) || ((s_recv_bytes == (-1)) && (errno != EINTR))) { /* disconnected by server */
            mzproxy_log_printf(def_mzproxy_log_level_info, "disconnected by server (id=%d, rx=%d)", (int)s_session->accept_socket, (int)s_recv_bytes);
            s_session->flags |= def_mzproxy_session_flag_disconnected_server;

            return(-1);
        }

        s_session->data_time_stamp = s_control->time_stamp;
        mzproxy_log_printf(def_mzproxy_log_level_info, "from server (id=%d, rx=%d)", (int)s_session->accept_socket, (int)s_recv_bytes);
#if __def_mzproxy_rx_dump__ != (0L)
        (void)mzproxy_log_dump(def_mzproxy_log_level_debug, (const void *)mzproxy_peek_fifo((mzproxy_fifo_t *)(&s_session->fifo[1]), (size_t)0), s_recv_bytes);
#endif   

        mzproxy_push_fifo((mzproxy_fifo_t *)(&s_session->fifo[1]), (const void *)0, (size_t)s_recv_bytes);
    }

    return(0);
}

/* service function */
static int mzproxy_do(mzproxy_control_t *s_control)
{
    mzproxy_session_t *s_session, *s_next_session;
    struct timeval s_timeval = {1l, 0l};
 
    /* get time stamp */
    s_control->time_stamp = mzproxy_time_stamp_sec();

    /* register event */
    mzproxy_register_event(s_control);
 
    /* synchronous I/O multiplexing */
    s_control->events = select(s_control->max_fd + 1, (fd_set *)(&s_control->fd_set_rx), (fd_set *)(&s_control->fd_set_tx), (fd_set *)0, (struct timeval *)(&s_timeval));
    if(mzproxy_builtin_expect(s_control->events == (-1), 0)) {
        mzproxy_log_printf(def_mzproxy_log_level_error, "synchronous I/O multiplexing failed ! (maxfds=%d, select)", s_control->max_fd + 1);

        return(-1);
    }
    
    if(s_control->events == 0)return(0); /* timeout */

#if 0L /* DEBUG */
    mzproxy_log_printf(def_mzproxy_log_level_debug, "events=%d (maxfds=%d)", s_control->events, s_control->max_fd + 1);
#endif

    /* listen event check */
    mzproxy_check_listen(s_control);

    /* event check */
    s_session = s_control->session;
    while((s_control->events > 0) && (s_session != ((mzproxy_session_t *)0))) {
        s_next_session = s_session->next;
        
        if(mzproxy_to_server(s_control, s_session) == (-1)) {
            mzproxy_del_session(s_control, s_session);
            s_session = s_next_session;

            continue;
        }
        
        if(mzproxy_to_client(s_control, s_session) == (-1)) {
            mzproxy_del_session(s_control, s_session);
            s_session = s_next_session;

            continue;
        }
        
        if(mzproxy_from_client(s_control, s_session) == (-1)) {
            mzproxy_del_session(s_control, s_session);
            s_session = s_next_session;

            continue;
        }
        
        if(mzproxy_from_server(s_control, s_session) == (-1)) {
            mzproxy_del_session(s_control, s_session);
            s_session = s_next_session;

            continue;
        }

        s_session = s_next_session;
    }

    /* check remain events */
    if(s_control->events > 0)mzproxy_log_printf(def_mzproxy_log_level_warning, "remain events=%d", s_control->events);

    return(0);
}

mzproxy_options_t *mzproxy_init_options(mzproxy_options_t *s_options)
{
    static const mzproxy_options_t sg_default_options = {
        "mzproxy",                     /* application name */
        0,                             /* do help */
        0,                             /* verbose */
#if def_mzproxy_windows == (0L)
        0,                             /* daemonize */
#endif
        0,                             /* listen port */
        60ul,                          /* active timeout */
        60ul,                          /* connect timeout */
        (const char *)0,               /* connect address */
        0,                             /* connect port */
        0,                             /* data timeout */
#if __def_mzproxy_use_ifaddrs__ != (0L)
        (char *)0,                     /* interface_name */
#endif  
        (size_t)__def_mzproxy_default_fifo_size__, /* buffer size */
        0u,                            /* limit session count */
#if __def_mzproxy_use_inet6__ != (0L)
        0,                             /* use inet6 */
#endif  
        0,                             /* use xinetd */
#if defined(mzproxy_telnetd_info_t)
        0,                             /* use telnetd */
        "/etc/issue.net",              /* issue file */
        "/bin/sh",                     /* login program */
#endif
#if defined(mzproxy_socks_info_t)
        0,                             /* use socks protocol */
#endif
        (int *)0                       /* stop value ptr */
    };

    return((mzproxy_options_t *)memcpy((void *)s_options, (const void *)(&sg_default_options), sizeof(mzproxy_options_t)));
}

int mzproxy_main_process(mzproxy_options_t *s_options)
{
    int s_result = EXIT_FAILURE, s_check;
    mzproxy_control_t s_local_control, *s_control;

    /* handle mapping */
    s_control = (mzproxy_control_t *)(&s_local_control);
    s_control->stop_service = s_options->stop_value_ptr;
    s_control->options = s_options;
#if defined(mzproxy_telnetd_info_t)
    {
        mzproxy_const_ptr_with_ptr_t s_arg0;
 
        s_arg0.const_ptr = (const void *)s_control->options->login_program;

        s_control->login_program_argv[0] = (char *)s_arg0.ptr;
        s_control->login_program_argv[1] = (char *)0;
    }
#endif
    s_control->time_stamp = mzproxy_time_stamp_sec();

    /* open listen socket */
    if(s_control->options->use_xinetd == 0) {
#if __def_mzproxy_use_inet6__ != (0L)
        if(s_control->options->use_inet6 == 0)s_control->socket = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
        else s_control->socket = socket(PF_INET6, SOCK_STREAM, IPPROTO_TCP);
#else
        s_control->socket = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
#endif  
        if(mzproxy_builtin_expect(s_control->socket == ((mzproxy_socket_t)(-1)), 0)) {
            mzproxy_log_printf(def_mzproxy_log_level_error, "can not open listen socket !");

            return(EXIT_FAILURE);
        }
    }
    else s_control->socket = (mzproxy_socket_t)(-1);

    if(s_control->options->use_xinetd == 0) { /* standalone mode */
        int s_sa_family;
        struct sockaddr_in s_listen_socket_address;
#if __def_mzproxy_use_inet6__ != (0L)
        struct sockaddr_in6 s_listen_socket_address6;
#endif  
        struct sockaddr *s_listen_socket_address_ptr;
        mzproxy_socklen_t s_listen_socket_address_size;

        (void)mzproxy_reuse_socket(s_control->socket);

        /* set listen socket address */
#if __def_mzproxy_use_inet6__ != (0L)
        if(s_control->options->use_inet6 == 0) {
            s_sa_family = AF_INET;
            s_listen_socket_address.sin_family = s_sa_family;
            s_listen_socket_address.sin_addr.s_addr = htonl(INADDR_ANY);
            s_listen_socket_address.sin_port = htons(s_control->options->listen_port);
            (void)memset((void *)(&s_listen_socket_address.sin_zero[0]), 0, sizeof(s_listen_socket_address.sin_zero));
            s_listen_socket_address_ptr = (struct sockaddr *)(&s_listen_socket_address);
            s_listen_socket_address_size = (mzproxy_socklen_t)sizeof(s_listen_socket_address);
        }
        else {
            s_sa_family = AF_INET6;
            (void)memset((void *)(&s_listen_socket_address6), 0, sizeof(s_listen_socket_address6));
            s_listen_socket_address6.sin6_family = s_sa_family;
            s_listen_socket_address6.sin6_flowinfo = 0;
            s_listen_socket_address6.sin6_addr = in6addr_any;
            s_listen_socket_address6.sin6_port = htons(s_control->options->listen_port);
            s_listen_socket_address_ptr = (struct sockaddr *)(&s_listen_socket_address6);
            s_listen_socket_address_size = (mzproxy_socklen_t)sizeof(s_listen_socket_address6);
        }
#else
        s_sa_family = AF_INET;
        s_listen_socket_address.sin_family = s_sa_family;
        s_listen_socket_address.sin_addr.s_addr = htonl(INADDR_ANY);
        s_listen_socket_address.sin_port = htons(s_control->options->listen_port);
        (void)memset((void *)(&s_listen_socket_address.sin_zero[0]), 0, sizeof(s_listen_socket_address.sin_zero));
        s_listen_socket_address_ptr = (struct sockaddr *)(&s_listen_socket_address);
        s_listen_socket_address_size = (mzproxy_socklen_t)sizeof(s_listen_socket_address);
#endif  
        
        /* apply interface to listen socket address */
        s_check = 0;
#if __def_mzproxy_use_ifaddrs__ != (0L)
        if(s_control->options->interface_name != ((char *)0)) { /* get address of interface */
            struct ifaddrs *s_ifaddrs, *s_trace;
            
            s_check = (-1);

            if(getifaddrs((struct ifaddrs **)(&s_ifaddrs)) == 0) {
                if(s_ifaddrs != ((struct ifaddrs *)0)) {
                    s_trace = s_ifaddrs;
                    while(s_trace != ((struct ifaddrs *)0)) {
                        if((strcmp(s_control->options->interface_name, s_trace->ifa_name) == 0) && (s_trace->ifa_addr->sa_family == s_sa_family)) {
                            char s_address_string[ 64 ];
                            unsigned int s_address_port;

#if __def_mzproxy_use_inet6__ != (0L)
                            if(s_control->options->use_inet6 == 0) {
                                (void)mzproxy_inet_ntop(AF_INET, memcpy((void *)(&s_listen_socket_address.sin_addr), (void *)(&(((struct sockaddr_in *)s_trace->ifa_addr)->sin_addr)), sizeof(s_listen_socket_address.sin_addr)), (char *)(&s_address_string[0]), sizeof(s_address_string));
                                s_address_port = ntohs(s_listen_socket_address.sin_port);
                            }
                            else {
                                (void)mzproxy_inet_ntop(AF_INET6, memcpy((void *)(&s_listen_socket_address6.sin6_addr), (void *)(&(((struct sockaddr_in6 *)s_trace->ifa_addr)->sin6_addr)), sizeof(s_listen_socket_address6.sin6_addr)), (char *)(&s_address_string[0]), sizeof(s_address_string));
                                s_address_port = ntohs(s_listen_socket_address6.sin6_port);
                            }
#else
                            (void)mzproxy_inet_ntop(AF_INET, memcpy((void *)(&s_listen_socket_address.sin_addr), (void *)(&(((struct sockaddr_in *)s_trace->ifa_addr)->sin_addr)), sizeof(s_listen_socket_address.sin_addr)), (char *)(&s_address_string[0]), sizeof(s_address_string));
                            s_address_port = ntohs(s_listen_socket_address.sin_port);
#endif       

                            mzproxy_log_printf(def_mzproxy_log_level_info, "interface %s binding (%s:%u)", s_control->options->interface_name, (char *)(&s_address_string[0]), s_address_port);

                            s_check = 0;

                            break;
                        }

                        s_trace = s_trace->ifa_next;
                    }

                    freeifaddrs(s_ifaddrs);
                }
            }
        }
#endif  
        
        if(s_check == 0) {
            if(s_control->options->use_xinetd == 0) {
                s_check = bind(s_control->socket, s_listen_socket_address_ptr, s_listen_socket_address_size);
                if(mzproxy_builtin_expect(s_check == (-1), 0)) {
                    mzproxy_log_printf(def_mzproxy_log_level_error, "bind failed ! (listen_socket=%d)", (int)s_control->socket);
                }
                else {
                    s_check = listen(s_control->socket, SOMAXCONN);
                    if(mzproxy_builtin_expect(s_check == (-1), 0)) {
                        mzproxy_log_printf(def_mzproxy_log_level_error, "listen failed ! (listen_socket=%d)", (int)s_control->socket);
                    }
                }
            }
        }
    }
    else s_check = 0;

    if(mzproxy_builtin_expect(s_check != 0, 0)) {
        mzproxy_log_puts(def_mzproxy_log_level_error, "not found interface");
    }
    else {
#if def_mzproxy_windows == (0L)
        if(s_control->options->daemonize != 0) {
            s_check = daemon(0, 1); /* daemonize */
            
            /* console log off */
            mzproxy_log_level(def_mzproxy_log_flag_syslog | (mzproxy_log_level(def_mzproxy_log_flag_get) & def_mzproxy_log_level_mask));
        }
        else s_check = 0;
#endif  

        if(mzproxy_builtin_expect(s_check != 0, 0)) {
            mzproxy_log_puts(def_mzproxy_log_level_error, "daemonize failed !");
        }
        else {
            int s_bye = 0;

            /* init session */
            s_control->session = (mzproxy_session_t *)0;
            s_control->session_count = 0u;
            
            s_result = EXIT_SUCCESS;

            if(s_control->options->use_xinetd == 0) { /* standalone mode */
                do { /* service loop */
                    if(mzproxy_do(s_control) == (-1))s_bye = 1; /* do service */
                    else if(s_control->stop_service != ((int *)0))s_bye = *s_control->stop_service;
                }while(s_bye == 0);
            }
            else { /* xinetd mode */
                (void)mzproxy_new_session(s_control, 0); /* xinetd socket */
                do {
                    if(s_control->session == ((mzproxy_session_t *)0))break;
                    if(mzproxy_do(s_control) == (-1))s_bye = 1; /* do service */
                    else if(s_control->stop_service != ((int *)0))s_bye = *s_control->stop_service;
                }while(s_bye == 0);
            }

            s_control->session = mzproxy_free_session(s_control, s_control->session, 0);
        }
    }
    
    mzproxy_closesocket(s_control->socket);

    return(s_result);
}

#endif

/* vim: set expandtab: */
/* End of source */
