/*
 Copyright (C) JAEHYUK CHO
 All rights reserved.
 
 Author: JAEHYUK CHO <minzkn@minzkn.com>
 
 http://www.minzkn.com
*/

#if !defined(__def_mzproxy_source_mzproxy_fifo_c__)
# define __def_mzproxy_source_mzproxy_fifo_c__ "mzproxy_fifo.c"

#include "mzproxy.h"

def_mzproxy_export mzproxy_shift_fifo_t *mzproxy_new_shift_fifo(mzproxy_shift_fifo_t *s_fifo, void *s_data, size_t s_size);
def_mzproxy_export void mzproxy_delete_shift_fifo(mzproxy_shift_fifo_t *s_fifo);
def_mzproxy_export size_t mzproxy_get_size_shift_fifo(mzproxy_shift_fifo_t *s_fifo);
def_mzproxy_export size_t mzproxy_set_size_shift_fifo(mzproxy_shift_fifo_t *s_fifo, size_t s_size);
def_mzproxy_export void *mzproxy_peek_shift_fifo(mzproxy_shift_fifo_t *s_fifo, size_t s_offset);
def_mzproxy_export ssize_t mzproxy_push_shift_fifo(mzproxy_shift_fifo_t *s_fifo, const void *s_data, size_t s_size);
def_mzproxy_export ssize_t mzproxy_pop_shift_fifo(mzproxy_shift_fifo_t *s_fifo, void *s_data, size_t s_size);

mzproxy_shift_fifo_t *mzproxy_new_shift_fifo(mzproxy_shift_fifo_t *s_fifo, void *s_data, size_t s_size)
{
    s_fifo->size = s_size;
    if(s_data == ((void *)0)) {
        s_fifo->alloc_data = malloc(s_fifo->size);
        if(s_fifo->alloc_data == ((void *)0)) {
            return((mzproxy_shift_fifo_t *)0);
        }
        s_fifo->data = (unsigned char *)s_fifo->alloc_data;
    }
    else {
        s_fifo->alloc_data = (void *)0;
        s_fifo->data = s_data;
    }
    s_fifo->entry = (size_t)0;

    return(s_fifo);
}

void mzproxy_delete_shift_fifo(mzproxy_shift_fifo_t *s_fifo)
{
    s_fifo->entry = (size_t)0;
    s_fifo->data = (void *)0;
    if(s_fifo->alloc_data != ((void *)0)) {
        free(s_fifo->alloc_data);
        s_fifo->alloc_data = (void *)0;
    }
    s_fifo->size = (size_t)0;
}

size_t mzproxy_get_size_shift_fifo(mzproxy_shift_fifo_t *s_fifo)
{
    return(s_fifo->entry);
}

size_t mzproxy_set_size_shift_fifo(mzproxy_shift_fifo_t *s_fifo, size_t s_size)
{
    s_fifo->entry = s_size;
    return(s_fifo->entry);
}

void *mzproxy_peek_shift_fifo(mzproxy_shift_fifo_t *s_fifo, size_t s_offset)
{
    return((void *)(&s_fifo->data[s_offset]));
}

ssize_t mzproxy_push_shift_fifo(mzproxy_shift_fifo_t *s_fifo, const void *s_data, size_t s_size)
{
    size_t s_copy_size, s_remain_size;

    s_remain_size = s_fifo->size - s_fifo->entry;
    s_copy_size = (s_remain_size < s_size) ? s_remain_size : s_size;
 
    if(s_copy_size > ((size_t)0)) {
        if(s_data != ((const void *)0)) {
            (void)memcpy((void *)(&s_fifo->data[s_fifo->entry]), s_data, s_copy_size);
        }
        s_fifo->entry += s_copy_size;
    }

    return((ssize_t)s_copy_size);
}

ssize_t mzproxy_pop_shift_fifo(mzproxy_shift_fifo_t *s_fifo, void *s_data, size_t s_size)
{
    size_t s_copy_size;

    s_copy_size = (s_fifo->entry < s_size) ? s_fifo->entry : s_size;

    if(s_copy_size > ((size_t)0)) {
        if(s_data != ((void *)0)) {
            (void)memcpy(s_data, (const void *)(&s_fifo->data[0]), s_copy_size);
        }
        
        if(s_fifo->entry == s_copy_size) { /* no shift */
            s_fifo->entry = (size_t)0;
        }
        else { /* do shift */
            (void)memmove((void *)(&s_fifo->data[0]), (const void *)(&s_fifo->data[s_copy_size]), s_fifo->entry - s_copy_size);
            s_fifo->entry -= s_copy_size;
        }
    }

    return(s_copy_size);
}

#endif

/* vim: set expandtab: */
/* End of source */
