/*
 Copyright (C) JAEHYUK CHO
 All rights reserved.
 
 Author: JAEHYUK CHO <minzkn@minzkn.com>
 
 http://www.minzkn.com
*/

#if !defined(__def_mzproxy_source_mzproxy_log_c__)
# define __def_mzproxy_source_mzproxy_log_c__ "mzproxy_log.c"

#include "mzproxy.h"

#if !defined(STDIN_FILENO)
# define STDIN_FILENO                                                (0)
#endif

def_mzproxy_export void mzproxy_use_log(void);
def_mzproxy_export void mzproxy_unuse_log(void);

def_mzproxy_export int mzproxy_log_level(int s_log_level);

def_mzproxy_export void mzproxy_log_puts(int s_log_level, const char *s_log_message);
def_mzproxy_export void mzproxy_log_vprintf(int s_log_level, const char *s_format, va_list s_var);
def_mzproxy_export void mzproxy_log_printf(int s_log_level, const char *s_format, ...) mzproxy_printf_varg_check(2,3);
def_mzproxy_export void mzproxy_log_dump(int s_log_level, const void *s_data, size_t s_size);

static int g_mzproxy_init_log = 0, g_mzproxy_log_level = def_mzproxy_log_flag_none | def_mzproxy_log_level_notice;

void mzproxy_use_log(void)
{
#if __def_mzproxy_use_syslog__ != (0L)
    openlog("mzproxy", LOG_PID, LOG_DAEMON);
#endif

    ++g_mzproxy_init_log;
}

void mzproxy_unuse_log(void)
{
    --g_mzproxy_init_log;

#if __def_mzproxy_use_syslog__ != (0L)
    closelog();
#endif
}

#if !defined(STDOUT_FILENO)
# define STDOUT_FILENO 1
#endif

int mzproxy_log_level(int s_log_level)
{
    if((s_log_level & def_mzproxy_log_flag_get) == def_mzproxy_log_flag_none) {
        if((s_log_level & def_mzproxy_log_flag_console) == def_mzproxy_log_flag_console) {
	
#if def_mzproxy_windows != 0L
            if(_isatty(STDOUT_FILENO) == 0) {
                /* can not use tty */
                s_log_level &= (~def_mzproxy_log_flag_console);
            }
#else
            if(isatty(1) == 0) {
                /* can not use tty */
                s_log_level &= (~def_mzproxy_log_flag_console);
            }
#endif	    
        }
        g_mzproxy_log_level = s_log_level;
    }

    return(g_mzproxy_log_level);
}

void mzproxy_log_puts(int s_log_level, const char *s_log_message)
{
    if((s_log_level & def_mzproxy_log_level_mask) > (g_mzproxy_log_level & def_mzproxy_log_level_mask)) {
        return;
    }
    
    if((g_mzproxy_log_level & def_mzproxy_log_flag_console) == def_mzproxy_log_flag_console) {
        (void)fputs(s_log_message, stdout);
        (void)fputs("\n", stdout);
    }

#if __def_mzproxy_use_syslog__ != (0L)
    if((g_mzproxy_init_log > 0) && ((g_mzproxy_log_level & def_mzproxy_log_flag_syslog) == def_mzproxy_log_flag_syslog) && 
       ((s_log_level & def_mzproxy_log_level_mask) != def_mzproxy_log_level_notice) && ((s_log_level & def_mzproxy_log_level_mask) <= def_mzproxy_log_level_info)) {
        static const int s_syslog_priority[def_mzproxy_log_level_mask + 1] = {
            /* 0x00 */ LOG_EMERG,
            /* 0x01 */ LOG_ALERT,
            /* 0x02 */ LOG_CRIT,
            /* 0x03 */ LOG_ERR,
            /* 0x04 */ LOG_WARNING,
            /* 0x05 */ LOG_NOTICE,
            /* 0x06 */ LOG_INFO,
            /* 0x07 */ LOG_DEBUG,
            0
        };
        
        syslog(s_syslog_priority[s_log_level & def_mzproxy_log_level_mask], "%s", s_log_message);
    }
#endif    
}

#define def_mzproxy_log_vprintf_buffer_size                          (32 << 10)
void mzproxy_log_vprintf(int s_log_level, const char *s_format, va_list s_var)
{
    char s_string[ def_mzproxy_log_vprintf_buffer_size ];
    
    if((s_log_level & def_mzproxy_log_level_mask) > (g_mzproxy_log_level & def_mzproxy_log_level_mask)) {
        return;
    }

    /* IMPORTANT: buffer overflow */
    (void)vsprintf((char *)(&s_string[0]), s_format, s_var);

    mzproxy_log_puts(s_log_level, (const char *)(&s_string[0]));
}

void mzproxy_log_printf(int s_log_level, const char *s_format, ...)
{
    va_list s_var;

    if((s_log_level & def_mzproxy_log_level_mask) > (g_mzproxy_log_level & def_mzproxy_log_level_mask)) {
        return;
    }

    va_start(s_var, s_format);
    mzproxy_log_vprintf(s_log_level, s_format, s_var);
    va_end(s_var);
}

#define def_mzproxy_dump_width                                       (16)
void mzproxy_log_dump(int s_log_level, const void *s_data, size_t s_size)
{
    size_t s_width, s_offset, s_local_offset;
    char s_print_buffer[ 8 + (3 * def_mzproxy_dump_width) + 3 + 1 + 1 + def_mzproxy_dump_width + 1 + 1];
    size_t s_print_buffer_offset;
    unsigned char s_unit_data[ def_mzproxy_dump_width + 1 ];
    
    if((s_log_level & def_mzproxy_log_level_mask) > (g_mzproxy_log_level & def_mzproxy_log_level_mask)) {
        return;
    }
    
    s_unit_data[ def_mzproxy_dump_width ] = (unsigned char)0u;
    s_offset = (size_t)0;
    while(s_offset < s_size) {
        s_width = ((s_size - s_offset) < def_mzproxy_dump_width) ? (s_size - s_offset) : def_mzproxy_dump_width;
        s_print_buffer_offset = sprintf((char *)(&s_print_buffer[0]), "%08lX", (unsigned long)s_offset);
        for(s_local_offset = ((size_t)0);s_local_offset < s_width;s_local_offset++) {
            s_unit_data[s_local_offset] = *(((const unsigned char *)s_data) + (s_offset + s_local_offset));
            s_print_buffer_offset += sprintf((char *)(&s_print_buffer[s_print_buffer_offset]), "%s%02X", (s_local_offset == ((size_t)(def_mzproxy_dump_width >> 1))) ? " | " : " ", (unsigned int)s_unit_data[s_local_offset]);
            if((s_unit_data[s_local_offset] & ((unsigned char)0x80)) || (s_unit_data[s_local_offset] < ((unsigned char)0x20)) || (s_unit_data[s_local_offset] == ((unsigned char)0x7f))) {
                s_unit_data[s_local_offset] = (unsigned char)'.';
            }
        }
        while(s_local_offset < ((size_t)def_mzproxy_dump_width)) {
            s_print_buffer_offset += sprintf((char *)(&s_print_buffer[s_print_buffer_offset]), (s_local_offset == ((size_t)(def_mzproxy_dump_width >> 1))) ? "     " : "   ");
            s_unit_data[s_local_offset] = (unsigned char)' ';
            s_local_offset++; 
        }
        s_print_buffer_offset += sprintf((char *)(&s_print_buffer[s_print_buffer_offset]), " [%s]", (char *)(&s_unit_data[0]));
        mzproxy_log_puts(s_log_level, (const char *)(&s_print_buffer[0]));
        s_offset += (size_t)def_mzproxy_dump_width;
    }
}

#endif

/* vim: set expandtab: */
/* End of source */
