/*
 Copyright (C) JAEHYUK CHO
 All rights reserved.
 
 Author: JAEHYUK CHO <minzkn@minzkn.com>
 
 http://www.minzkn.com
*/

#if !defined(__def_mzproxy_source_mzproxy_session_c__)
# define __def_mzproxy_source_mzproxy_session_c__ "mzproxy_session.c"

#if !defined(_ISOC99_SOURCE)
# define _ISOC99_SOURCE (1L)
#endif

#if !defined(_GNU_SOURCE)
# define _GNU_SOURCE (1L)
#endif

#include "mzproxy.h"

def_mzproxy_export mzproxy_session_t * mzproxy_free_session(mzproxy_control_t *s_control, mzproxy_session_t *s_session, int s_nokill);
def_mzproxy_export mzproxy_session_t * mzproxy_new_session(mzproxy_control_t *s_control, mzproxy_socket_t s_accept_socket);
def_mzproxy_export void mzproxy_del_session(mzproxy_control_t *s_control, mzproxy_session_t *s_session);

mzproxy_session_t * mzproxy_free_session(mzproxy_control_t *s_control, mzproxy_session_t *s_session, int s_nokill)
{
    mzproxy_session_t *s_prev_session;
    int s_index;

    (void)s_control;

    while(s_session != ((mzproxy_session_t *)0)) {
        s_prev_session = s_session;
        s_session = s_session->next;

        for(s_index = 0;s_index < __def_mzproxy_fifo_count__;s_index++) {
            if(mzproxy_get_size_fifo((mzproxy_fifo_t *)(&s_prev_session->fifo[s_index])) > ((size_t)0)) {
                mzproxy_log_printf(def_mzproxy_log_level_warning, "drop fifo (id=%d, drop_fifo_size[%d]=%lu)", (int)s_prev_session->accept_socket, (int)s_index, (unsigned long)mzproxy_get_size_fifo((mzproxy_fifo_t *)(&s_prev_session->fifo[s_index])));
            }
            mzproxy_delete_fifo((mzproxy_fifo_t *)(&s_prev_session->fifo[s_index]));
        }

#if defined(mzproxy_telnetd_info_t)
        if(s_prev_session->telnetd_info != ((mzproxy_telnetd_info_t *)0)) {
            if((s_nokill == 0) && (s_prev_session->telnetd_info->pid != ((pid_t)(-1))) && (s_prev_session->telnetd_info->pid != ((pid_t)0))) {
                /* force kill */
                if(mzproxy_builtin_expect(kill(s_prev_session->telnetd_info->pid, SIGKILL) == (-1), 0)) {
                    mzproxy_log_printf(def_mzproxy_log_level_error, "process kill failed ! (pid=%d)", (int)s_prev_session->telnetd_info->pid);
                }

                /* wait for login program */
                if(mzproxy_builtin_expect(waitpid(s_prev_session->telnetd_info->pid, (int *)0, 0) == ((pid_t)(-1)), 0)) {
                    mzproxy_log_printf(def_mzproxy_log_level_error, "waitpid failed ! (pid=%d)", (int)s_prev_session->telnetd_info->pid);
                }
            }
        }
#endif  
        
        if(s_prev_session->connect_address != ((char *)0))free((void *)s_prev_session->connect_address);

        if(s_prev_session->accept_socket != ((mzproxy_socket_t)(-1)))mzproxy_closesocket(s_prev_session->accept_socket);

        if(s_prev_session->client_socket != ((mzproxy_socket_t)(-1))) {
#if defined(mzproxy_telnetd_info_t)
            if(s_prev_session->telnetd_info != ((mzproxy_telnetd_info_t *)0)) {
# if defined(EINTR)                
                while(mzproxy_builtin_expect((close((int)s_prev_session->client_socket) == (-1)) && (errno == EINTR), 0));
# else
                (void)close((int)s_prev_session->client_socket);
# endif
            }
            else {
               mzproxy_closesocket(s_prev_session->client_socket);
            }
#else            
            mzproxy_closesocket(s_prev_session->client_socket);
#endif   
        }

        free((void *)s_prev_session);
    }

    return((mzproxy_session_t *)0);
}

mzproxy_session_t * mzproxy_new_session(mzproxy_control_t *s_control, mzproxy_socket_t s_accept_socket)
{
    size_t s_session_info_size;
    mzproxy_session_t *s_session;
    int s_index;

    s_session_info_size = sizeof(mzproxy_session_t);
#if defined(mzproxy_telnetd_info_t)
    if(s_control->options->use_telnetd != 0) {
        s_session_info_size += sizeof(mzproxy_telnetd_info_t);
    }
#endif
#if defined(mzproxy_socks_info_t)
    if(s_control->options->use_socks != 0) {
        s_session_info_size += sizeof(mzproxy_socks_info_t);
    }
#endif

    s_session = (mzproxy_session_t *)malloc(s_session_info_size);
    if(mzproxy_builtin_expect(s_session == ((mzproxy_session_t *)0), 0)) {
        mzproxy_closesocket(s_accept_socket);

        return((mzproxy_session_t *)0);
    }
    (void)memset((void *)s_session, 0, s_session_info_size);

    (void)mzproxy_tos_lowdelay_socket(s_accept_socket);

    (void)mzproxy_oobinline_socket(s_accept_socket);

    /* s_session->prev = (mzproxy_session_t *)0; */
    /* s_session->next = (mzproxy_session_t *)0; */
    s_session->connect_time_stamp = s_control->time_stamp;
    s_session->data_time_stamp = s_control->time_stamp;
    s_session->flags = def_mzproxy_session_flag_none | def_mzproxy_session_flag_active_connect | def_mzproxy_session_flag_active_server; 
    s_session->accept_socket = s_accept_socket;
    s_session->client_socket = (mzproxy_socket_t)(-1);
    /* s_session->connect_address = (char *)0; */
    s_session->connect_port = s_control->options->connect_port;
    /* (void)memset((void *)(&s_session->fifo[0]), 0, sizeof(s_session->fifo)); */

    s_session_info_size = sizeof(mzproxy_session_t);
#if defined(mzproxy_telnetd_info_t)
    /* s_session->telnetd_info = (mzproxy_telnetd_info_t *)0; */
    if(s_control->options->use_telnetd != 0) {
        s_session->telnetd_info = (mzproxy_telnetd_info_t *)(((unsigned char *)s_session) + s_session_info_size);
        s_session_info_size += sizeof(mzproxy_telnetd_info_t);
        s_session->telnetd_info->pid = (pid_t)(-1);
    }
#endif
#if defined(mzproxy_socks_info_t)
    /* s_session->socks_info = (mzproxy_socks_info_t *)0; */
    if(s_control->options->use_socks != 0) {
        s_session->socks_info = (mzproxy_socks_info_t *)(((unsigned char *)s_session) + s_session_info_size);
        s_session_info_size += sizeof(mzproxy_socks_info_t);
        s_session->socks_info->flags = def_mzproxy_socks_flag_none;
        /* (void)memset((void *)(&s_session->socks_info->select_method[0]), 0, sizeof(s_session->socks_info->select_method)); */
        /* s_session->socks_info->done_methods = 0; */
    }
#endif

    for(s_index = 0;s_index < __def_mzproxy_fifo_count__;s_index++) {
        if(mzproxy_new_fifo((mzproxy_fifo_t *)(&s_session->fifo[s_index]), (void *)0, s_control->options->fifo_size) == ((mzproxy_fifo_t *)0)) {
            mzproxy_log_printf(def_mzproxy_log_level_error, "not enough memory ! (id=%d, session fifo)", (int)s_session->accept_socket);

            return(mzproxy_free_session(s_control, s_session, 0));
        }
    }

#if defined(mzproxy_telnetd_info_t)
    if(s_session->telnetd_info != ((mzproxy_telnetd_info_t *)0)) { /* create pt handle */
        /* default IAC session start */
        static const unsigned char scg_mzproxy_default_iac[] = {
            IAC, DO, TELOPT_ECHO,
            IAC, DO, TELOPT_NAWS,
            IAC, DO, TELOPT_LFLOW, /* for winsize */
            IAC, WILL, TELOPT_ECHO,
            IAC, WILL, TELOPT_SGA
        };
        char *s_pts_name; 
    
        /* default iac code size to pt fifo */
        mzproxy_push_fifo((mzproxy_fifo_t *)(&s_session->fifo[1]), (const void *)(&scg_mzproxy_default_iac[0]), sizeof(scg_mzproxy_default_iac));
        
        /* open the pseudo-terminal master (PTM) */
        s_session->client_socket = (mzproxy_socket_t)getpt();
        if(mzproxy_builtin_expect(s_session->client_socket == ((mzproxy_socket_t)(-1)), 0)) {
             mzproxy_log_printf(def_mzproxy_log_level_error, "getpt failed ! (id=%d)", (int)s_session->accept_socket);
            s_session->flags |= def_mzproxy_session_flag_disconnected_server;

            return(mzproxy_free_session(s_control, s_session, 0));
        }

        s_session->flags |= def_mzproxy_session_flag_connecting;

        /* grant access to the slave pseudo-terminal */
        if(mzproxy_builtin_expect(grantpt((int)s_session->client_socket) == (-1), 0)) {
            mzproxy_log_printf(def_mzproxy_log_level_error, "grantpt failed ! (id=%d)", (int)s_session->accept_socket);

            return(mzproxy_free_session(s_control, s_session, 0));
        }

        /* unlock a pseudo-terminal master/slave pair */
        if(mzproxy_builtin_expect(unlockpt((int)s_session->client_socket) == (-1), 0)) {
            mzproxy_log_printf(def_mzproxy_log_level_error, "unlockpt failed ! (id=%d)", (int)s_session->accept_socket);

            return(mzproxy_free_session(s_control, s_session, 0));
        }

        /* get pts name */
        s_pts_name = strdup(ptsname((int)s_session->client_socket));
        if(mzproxy_builtin_expect(s_pts_name == ((char *)0), 0)) {
            mzproxy_log_printf(def_mzproxy_log_level_error, "not enough memory ! (id=%d, pts_name)", (int)s_session->accept_socket);

            return(mzproxy_free_session(s_control, s_session, 0));
        }
        mzproxy_log_printf(def_mzproxy_log_level_info, "new pts name is \"%s\".", s_pts_name);
 
        s_session->telnetd_info->pid = fork(); /* process fork */
        if(mzproxy_builtin_expect(s_session->telnetd_info->pid == ((pid_t)(-1)), 0)) {
            mzproxy_log_printf(def_mzproxy_log_level_error, "fork failed ! (id=%d)", (int)s_session->accept_socket);
            free((void *)s_pts_name);

            return(mzproxy_free_session(s_control, s_session, 0));
        }

        if(s_session->telnetd_info->pid == ((pid_t)0)) {
            /* close all session (close handle) */
            s_control->session = mzproxy_free_session(s_control, s_control->session, 1);
            (void)mzproxy_free_session(s_control, s_session, 1); /* current session */
            (void)mzproxy_closesocket(s_control->socket); /* close listen socket */

            /* close stdin, stdout, stderr */
            (void)close(STDERR_FILENO); (void)close(STDOUT_FILENO); (void)close(STDIN_FILENO);

            if(setsid() != (-1)) {
                int s_tty_handle = open(s_pts_name, O_RDWR | O_NOCTTY); /* #0 open stdin */

                if(s_tty_handle != (-1)) {
                    int s_stdout_handle, s_stderr_handle;
                    struct termios s_termios;

                    s_stdout_handle = dup(s_tty_handle); /* #1 copy to stdout */
                    s_stderr_handle = dup(s_tty_handle); /* #2 copy to stderr */

                    (void)tcsetpgrp(0, getpid());
                    (void)ioctl(s_tty_handle, TIOCSCTTY, NULL);

                    if(tcgetattr(s_tty_handle, (struct termios *)(&s_termios)) == 0) { /* set pt property */
                        s_termios.c_iflag |= ICRNL;
                        s_termios.c_iflag &= ~(IXOFF);
                        s_termios.c_oflag |= XTABS;
                        s_termios.c_oflag |= ONLCR;
#if 0L /* TRY: enable implementation-defined output processing (SeeAlso: man termios) */
                        s_termios.c_oflag |= OPOST;
#endif      
                        s_termios.c_lflag |= ECHO;
                        (void)tcsetattr(s_tty_handle, TCSANOW, (struct termios *)(&s_termios));
                    }
                    if(s_control->options->issue_file != ((char *)0)) { /* issue file to pt */
                        int s_issue_handle = open(s_control->options->issue_file, O_RDONLY);
                        ssize_t s_read_bytes, s_write_bytes;
                        unsigned char s_byte, s_cr = (unsigned char)'\r';

                        if(s_issue_handle != (-1)) {
                            for(;;) {
			        do {
                                    s_read_bytes = read(s_issue_handle, (void *)(&s_byte), sizeof(s_byte));
				}while((s_read_bytes == ((ssize_t)(-1))) && (errno == EINTR));
                                if(s_read_bytes != (sizeof(s_byte)))break;

                                if(s_byte == '\n') {
				    do {
				        s_write_bytes = write(s_stdout_handle, (void *)(&s_cr), sizeof(s_cr)); /* <LF> to <CR><LF> */
				    }while((s_write_bytes == ((ssize_t)(-1))) && (errno == EINTR));
				}
				do {
                                    s_write_bytes = write(s_stdout_handle, (void *)(&s_byte), sizeof(s_cr));
				}while((s_write_bytes == ((ssize_t)(-1))) && (errno == EINTR));
                            }

                            (void)close(s_issue_handle);
                        }
                    }

                    if(execv(s_control->options->login_program, s_control->login_program_argv) == (-1)) {
                        mzproxy_log_printf(def_mzproxy_log_level_error, "execv failed ! (login_program=\"%s\")", (const char *)s_control->options->login_program);
                    }

                    (void)close(s_tty_handle);
                }
            }
            exit(EXIT_FAILURE);
        }
        
        s_session->flags &= (~def_mzproxy_session_flag_connecting);
        s_session->flags |= def_mzproxy_session_flag_connected;

        free((void *)s_pts_name);
    }
    else
#endif
    {
        s_session->client_socket = socket(
#if __def_mzproxy_use_inet6__ != (0L)
            (s_control->options->use_inet6 == 0) ? PF_INET : PF_INET6, SOCK_STREAM,
#else
            PF_INET,
#endif
            IPPROTO_TCP);
        if(mzproxy_builtin_expect(s_session->client_socket == ((mzproxy_socket_t)(-1)), 0)) {
            mzproxy_log_printf(def_mzproxy_log_level_error, "can not open socket ! (id=%d, client_socket)", (int)s_session->accept_socket);

            return(mzproxy_free_session(s_control, s_session, 0));
        }

        (void)mzproxy_linger_on(s_session->client_socket);

#if defined(mzproxy_socks_info_t)
        if(s_session->socks_info != ((mzproxy_socks_info_t *)0)) {
            s_session->flags &= (~(def_mzproxy_session_flag_active_connect | def_mzproxy_session_flag_active_server));
        }
#endif
    }
   
    /* add new session to head */
    if(s_control->session != ((mzproxy_session_t *)0))s_control->session->prev = s_session;
    s_session->next = s_control->session;
    s_control->session = s_session;
    s_control->session_count++;

    return(s_session);
}

void mzproxy_del_session(mzproxy_control_t *s_control, mzproxy_session_t *s_session)
{
    if(s_session == ((mzproxy_session_t *)0))return;

    if(s_session->next != ((mzproxy_session_t *)0))s_session->next->prev = s_session->prev;
    if(s_session->prev != ((mzproxy_session_t *)0))s_session->prev->next = s_session->next;
    else s_control->session = s_session->next;
    s_session->prev = (mzproxy_session_t *)0;
    s_session->next = (mzproxy_session_t *)0;

    (void)mzproxy_free_session(s_control, s_session, 0);

    s_control->session_count--;
}

#endif

/* vim: set expandtab: */
/* End of source */
