/*
 Copyright (C) JAEHYUK CHO
 All rights reserved.
 
 Author: JAEHYUK CHO <minzkn@minzkn.com>
 
 http://www.minzkn.com
*/

#if !defined(__def_mzproxy_source_mzproxy_socket_c__)
# define __def_mzproxy_source_mzproxy_socket_c__ "mzproxy_socket.c"

#if !defined(_ISOC99_SOURCE)
# define _ISOC99_SOURCE (1L)
#endif

#if !defined(_GNU_SOURCE)
# define _GNU_SOURCE (1L)
#endif

#include "mzproxy.h"

def_mzproxy_export int __mzproxy_init_network(int s_bool);
def_mzproxy_export void mzproxy_closesocket(mzproxy_socket_t s_socket);
def_mzproxy_export const char * mzproxy_inet_ntop(int s_family, const void *s_inX_addr, char *s_address, mzproxy_socklen_t s_address_length);
def_mzproxy_export int mzproxy_nonblock_socket(mzproxy_socket_t s_socket, int s_enable);
def_mzproxy_export int mzproxy_linger_on(mzproxy_socket_t s_socket);
def_mzproxy_export int mzproxy_reuse_socket(mzproxy_socket_t s_socket);
def_mzproxy_export int mzproxy_oobinline_socket(mzproxy_socket_t s_socket);
def_mzproxy_export int mzproxy_tos_lowdelay_socket(mzproxy_socket_t s_socket);
def_mzproxy_export ssize_t mzproxy_send(mzproxy_socket_t s_socket, const void *s_data, size_t s_size);
def_mzproxy_export ssize_t mzproxy_recv(mzproxy_socket_t s_socket, void *s_data, size_t s_size);

int __mzproxy_init_network(int s_bool)
{
    static volatile int sg_init_count = 0;
    int s_init_count;

    if(s_bool != 0) {
        s_init_count = sg_init_count;
        sg_init_count++;

        if(mzproxy_builtin_expect(s_init_count < 0, 0)) {
            return(-1);
        }

        if(s_init_count > 0) {
            return(1);
        }

#if def_mzproxy_winsock != (0L)
        { /* initialize windows socket v2.2 (required DLL resource : "WS2_32.DLL") */
            BYTE s_require_major_version;
            BYTE s_require_minor_version;
            WORD s_require_version;
            WSADATA s_wsadata;

            s_require_major_version = (BYTE)2;
            s_require_minor_version = (BYTE)2;
            s_require_version = MAKEWORD(/* low */ s_require_major_version, /* high */ s_require_minor_version);
           
            /* int WSAStartup(WORD wVersionRequested, LPWSADATA lpWSAData); */
            if(mzproxy_builtin_expect(WSAStartup(s_require_version, (LPWSADATA)(&s_wsadata)) != 0, 0)) {
                return(0);
            }

            if((LOBYTE(s_wsadata.wVersion) == s_require_major_version) &&
               (HIBYTE(s_wsadata.wVersion) == s_require_minor_version)) {
                return(1);
            }

            mzproxy_log_printf(def_mzproxy_log_level_error, "Windows Socket v%d.%d is not supported !", (int)(LOBYTE(s_wsadata.wVersion)), (int)(HIBYTE(s_wsadata.wVersion)));

            /* int WSACleanup(void); */
            (void)WSACleanup();
        }

        return(0);
#else
        return(1);
#endif
    }

    sg_init_count--;
    s_init_count = sg_init_count;

    if(mzproxy_builtin_expect(s_init_count < 0, 0)) {
        return(-1);
    }
        
    if(s_init_count > 0) {
        return(1);
    }

#if def_mzproxy_winsock != (0L)
    /* int WSACleanup(void); */
    (void)WSACleanup();

    return(1);
#else
    return(1);
#endif
}

void mzproxy_closesocket(mzproxy_socket_t s_socket)
{
#if def_mzproxy_winsock != (0L)
    closesocket(s_socket);
#else
# if defined(EINTR)
    while(mzproxy_builtin_expect((close(s_socket) == (-1)) && (errno == EINTR), 0));
# else
    (void)close(s_socket);
# endif
#endif
}

const char * mzproxy_inet_ntop(int s_family, const void *s_inX_addr, char *s_address, mzproxy_socklen_t s_address_length)
{
    if(mzproxy_builtin_expect((s_address == ((char *)0)) || (s_address_length <= ((mzproxy_socklen_t)0)), 0)) {
        return((const char *)0);
    }

#if defined(AF_INET)
    if(s_family == AF_INET) {
        const unsigned char *s_byte_array;

        s_byte_array = (const unsigned char *)s_inX_addr;

# if def_mzproxy_winsock != (0L)
        (void)sprintf(s_address, "%u.%u.%u.%u",
            (unsigned int)s_byte_array[0], (unsigned int)s_byte_array[1],
            (unsigned int)s_byte_array[2], (unsigned int)s_byte_array[3]
        );
# else
        (void)snprintf(s_address, (size_t)s_address_length, "%u.%u.%u.%u",
            (unsigned int)s_byte_array[0], (unsigned int)s_byte_array[1],
            (unsigned int)s_byte_array[2], (unsigned int)s_byte_array[3]
        );
# endif        
        
        return(s_address);
    }
#endif 

#if (__def_mzproxy_use_inet6__ != (0L)) && defined(AF_INET6)
    if(s_family == AF_INET6) {
        const struct in6_addr *s_in6_addr;
        unsigned short s_word_array[ 8 ];
        int s_offset, s_entry, s_count, s_max_entry, s_max_count;
        
        s_in6_addr = (const struct in6_addr *)s_inX_addr;

        s_entry = (-1);
        s_max_entry = (-1);
        s_max_count = 0;
        s_offset = 0;
        while(s_offset < 8) {
# if def_mzproxy_winsock != (0L)
            s_word_array[s_offset] = ntohs(s_in6_addr->s6_words[s_offset]);
# else
            s_word_array[s_offset] = ntohs(s_in6_addr->s6_addr16[s_offset]);
# endif
            if(s_word_array[s_offset] == ((unsigned short)0)) {
                s_entry = s_offset;
                s_count = 1;
                s_offset++;
                while(s_offset < 8) {
# if def_mzproxy_winsock != (0L)
                    s_word_array[s_offset] = ntohs(s_in6_addr->s6_words[s_offset]);
# else
                    s_word_array[s_offset] = ntohs(s_in6_addr->s6_addr16[s_offset]);
# endif
                    if(s_word_array[s_offset] != ((unsigned short)0)) {
                        break;
                    }
                    s_offset++;
                    s_count++;
                }
                
                if(s_count > s_max_count) {
                    s_max_entry = s_entry;
                    s_max_count = s_count;
                }
            }
            s_offset++;
        }
       
        s_address[0] = '\0';
        s_offset = 0;
        while(s_offset < 8) {
            if(s_offset == s_max_entry) {
                (void)strcpy((char *)(&s_address[strlen(s_address)]), ":");
                s_address_length--;
                s_offset += s_max_count;
                if(s_offset >= 8) {
                    (void)strcpy((char *)(&s_address[strlen(s_address)]), ":");
                    s_address_length--;
                }
            }
            else {
                if(s_offset > 0) {
                    (void)strcpy((char *)(&s_address[strlen(s_address)]), ":");
                    s_address_length--;
                }
# if def_mzproxy_winsock != (0L)
                s_address_length -= sprintf((char *)(&s_address[strlen(s_address)]), "%x", (unsigned int)s_word_array[s_offset]);
# else
                s_address_length -= snprintf((char *)(&s_address[strlen(s_address)]), (size_t)s_address_length, "%x", (unsigned int)s_word_array[s_offset]);
# endif
                s_offset++;   
            }
        }

        return(s_address);
    }
#endif

    return((const char *)memset((void *)s_address, 0, (size_t)s_address_length));
}

int mzproxy_nonblock_socket(mzproxy_socket_t s_socket, int s_enable)
{
#if def_mzproxy_winsock != (0L)
    long unsigned int s_flags = (s_enable == 0) ? 0ul : 1ul;

    return((ioctlsocket(s_socket, FIONBIO, &s_flags) == 0) ? 1 : (-1));
#else
    int s_flags;

    s_flags = fcntl(s_socket, F_GETFL);
    if(mzproxy_builtin_expect(s_flags == (-1), 0))return(-1);

    return((fcntl((int)s_socket, F_SETFL, (s_enable == 0) ? (s_flags & (~(O_NONBLOCK))) : (s_flags | (O_NONBLOCK))) == 0) ? 1 : (-1));
#endif 
}
   
int mzproxy_linger_on(mzproxy_socket_t s_socket)
{
    struct linger s_linger = {1, 0};

#if def_mzproxy_winsock != (0L)
    if(mzproxy_builtin_expect(setsockopt((int)s_socket, SOL_SOCKET, SO_LINGER, (char *)(&s_linger), (int)sizeof(s_linger)) == (-1), 0))return((-1));
#else    
    if(mzproxy_builtin_expect(setsockopt((int)s_socket, SOL_SOCKET, SO_LINGER, (void *)(&s_linger), (mzproxy_socklen_t)sizeof(s_linger)) == (-1), 0))return((-1));
#endif

    return(1);
}
        
int mzproxy_reuse_socket(mzproxy_socket_t s_socket)
{
    int s_reuse_on = 1;
    if(mzproxy_builtin_expect(setsockopt(s_socket, SOL_SOCKET, SO_REUSEADDR, (void *)(&s_reuse_on), (mzproxy_socklen_t)sizeof(s_reuse_on)) == (-1), 0)) {
        return(0);
    }

    return(1);
}        

int mzproxy_oobinline_socket(mzproxy_socket_t s_socket)
{
#if defined(SO_OOBINLINE)
    int s_oobinline_on = 1;
    if(mzproxy_builtin_expect(setsockopt(s_socket, SOL_SOCKET, SO_OOBINLINE, (void *)(&s_oobinline_on), (mzproxy_socklen_t)sizeof(s_oobinline_on)) == (-1), 0)) {
        return(0);
    }
    
    return(1);
#else    
    (void)s_socket;
    return(-1);
#endif   
}

int mzproxy_tos_lowdelay_socket(mzproxy_socket_t s_socket)
{
#if defined(IP_TOS) && defined(IPTOS_LOWDELAY)
    int s_tos = IPTOS_LOWDELAY;

    if(mzproxy_builtin_expect(setsockopt(s_socket, IPPROTO_IP, IPTOS_LOWDELAY, (void *)(&s_tos), (mzproxy_socklen_t)sizeof(s_tos)) == (-1), 0)) {
       return(0);
    }

    return(1);
#else
    (void)s_socket;
    return(-1);
#endif
}

ssize_t mzproxy_send(mzproxy_socket_t s_socket, const void *s_data, size_t s_size)
{
#if def_mzproxy_winsock != (0L)
    return((ssize_t)send(s_socket, s_data, (int)s_size, 0));
#else
    return(send(s_socket, s_data, s_size, MSG_NOSIGNAL));
#endif
}

ssize_t mzproxy_recv(mzproxy_socket_t s_socket, void *s_data, size_t s_size)
{
#if def_mzproxy_winsock != (0L)
    return((ssize_t)recv(s_socket, s_data, (int)s_size, 0));
#else
    return(recv(s_socket, s_data, s_size, MSG_NOSIGNAL));
#endif
}

#endif

/* vim: set expandtab: */
/* End of source */
